/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
==========================================================================

						 - SPLITMODELS -

==========================================================================
*/
// - Adding the Player models in split pieces
// by Jalisk0

#include "cg_local.h"
#include "../game/gs_pmodels.h"


//======================================================================
//						PlayerModel Registering
//======================================================================

pmodel_t		cg_entPModels[MAX_EDICTS];
pmodelinfo_t	cg_PModelInfos[CG_MAX_PMODELS];
pskin_t			cg_pSkins[CG_MAX_PSKINS];

static char *pmPartNames[] = { "lower", "upper", "head", NULL };
static char *pmTagNames[] = { "tag_torso", "tag_head", NULL };


/*
================
CG_PModelsInit

  link PModel handles to it's cg_entities
================
*/
void CG_PModelsInit(void)
{
	memset(cg_entPModels, 0, sizeof(cg_entPModels));
}


/*
================
CG_CopyAnimation
================
*/
void CG_CopyAnimation (pmodelinfo_t *pmodelinfo, int put, int pick)
{
	pmodelinfo->firstframe[put] = pmodelinfo->firstframe[pick];
	pmodelinfo->lastframe[put] = pmodelinfo->lastframe[pick];
	pmodelinfo->loopingframes[put] = pmodelinfo->loopingframes[pick];
	pmodelinfo->frametime[put] = pmodelinfo->frametime[pick];
}

/*
================
CG_ParseAnimationScript

  Reads the animation config file.

0 = first frame
1 = number of frames/lastframe
2 = looping frames
3 = frame time
  
Note: The animations count begins at 1, not 0. I preserve zero for "no animation change"
---------------
New keywords:
nooffset: Uses the first frame value at the script, and no offsets, for the animation.
alljumps: Uses 3 different jump animations (bunnyhoping)
islastframe: second value of each animation is lastframe instead of numframes
	
Q3 keywords:
sex m/f : sets gender
	  
Q3 Unsupported:
headoffset <value> <value> <value>
footsteps
================
*/
qboolean CG_ParseAnimationScript ( pmodelinfo_t *pmodelinfo, char *filename )
{
	qbyte		*buf;
	char		*ptr, *token;
	int			rounder, counter, i, offset;
	qboolean	nooffset = qfalse;
	qboolean	alljumps = qfalse;
	qboolean	islastframe = qfalse;
	qboolean	debug = qtrue;
	int			anim_data[4][PMODEL_MAX_ANIMS];
	int			filenum;
	int			length;

	
	pmodelinfo->sex = GENDER_MALE;
	rounder = 0;
	counter = 1; //reseve 0 for 'no animation'

	if (!cg_debugPlayerModels->integer)
		debug = qfalse;

	// load the file
	length = trap_FS_FOpenFile( filename, &filenum, FS_READ );
	if( length == -1 ) {
		CG_Printf ("%sCouldn't load animation script: %s%s\n", S_COLOR_YELLOW, filename, S_COLOR_WHITE);
		return qfalse;
	}
	if( !length ) {
		trap_FS_FCloseFile( filenum );
		return qfalse;
	}
	buf = CG_Malloc( length + 1 );
	trap_FS_Read( buf, length, filenum );
	trap_FS_FCloseFile( filenum );
	
	if( !buf ) {
		CG_Free(buf);
		CG_Printf ("%sCouldn't load animation script: %s%s\n",S_COLOR_YELLOW, filename, S_COLOR_WHITE);
		return qfalse;
	}

	if (debug)
		CG_Printf ("%sLoading animation script: %s%s\n",S_COLOR_BLUE, filename, S_COLOR_WHITE);
	
	//proceed
	ptr = ( char * )buf;
	while ( ptr )
	{
		token = COM_ParseExt ( &ptr, qtrue );
		if ( !token )
			break;
		
		if (*token < '0' || *token > '9') {
			
			//headoffset (ignored)
			if ( !Q_stricmp (token, "headoffset") ) 
			{
				if (debug)
					CG_Printf ("%sScript: Ignored: %s%s", S_COLOR_YELLOW, token, S_COLOR_WHITE);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				if (debug)
					CG_Printf ("%s%s - %s", S_COLOR_YELLOW, token, S_COLOR_WHITE);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				if (debug)
					CG_Printf ("%s%s - %s", S_COLOR_YELLOW, token, S_COLOR_WHITE);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				if (debug)
					CG_Printf ("%s%s %s\n", S_COLOR_YELLOW, token, S_COLOR_WHITE);
				
			//gender
			} else if ( !Q_stricmp (token, "sex") ) {
				
				if (debug)
					CG_Printf ("%sScript: %s:%s", S_COLOR_BLUE, token, S_COLOR_WHITE);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				
				if ( token[0] == 'm' || token[0] == 'M' ) {
					pmodelinfo->sex = GENDER_MALE;
						if (debug)
							CG_Printf ("%s%s -Gender set to MALE%s\n", S_COLOR_BLUE, token, S_COLOR_WHITE);
					
				} else if ( token[0] == 'f' || token[0] == 'F' ) {
					
					pmodelinfo->sex = GENDER_FEMALE;
						if (debug)
							CG_Printf ("%s%s -Gender set to FEMALE%s\n", S_COLOR_BLUE, token, S_COLOR_WHITE);
				} else {
					pmodelinfo->sex = GENDER_NEUTRAL;
						if (debug)
							CG_Printf ("%s%s -Gender set to NEUTRAL%s\n", S_COLOR_BLUE, token, S_COLOR_WHITE);
				}
				
			//nooffset
			} else if ( !Q_stricmp (token, "nooffset") ) {
				nooffset = qtrue;
				if (debug)
					CG_Printf ("%sScript: No offset on lower frames%s\n", S_COLOR_BLUE, S_COLOR_WHITE);

			//alljumps
			} else if ( !Q_stricmp (token, "alljumps") ) {
				alljumps = qtrue;
				if (debug)
					CG_Printf ("%sScript: Using all jump animations%s\n", S_COLOR_BLUE, S_COLOR_WHITE);

			//islastframe
			} else if ( !Q_stricmp (token, "islastframe") ) {
				islastframe = qtrue;
				if (debug)
					CG_Printf ("%sScript: Second value is read as lastframe%s\n", S_COLOR_BLUE, S_COLOR_WHITE);

			} else if (token[0] && debug)
				CG_Printf ("%sScript: ignored: %s%s\n", S_COLOR_YELLOW, token, S_COLOR_WHITE);
			
		} else {
			
			//frame & animation values
			i = (int)atoi(token);
			if (debug) {
				if(rounder == 0)
					CG_Printf ("%sScript: %s", S_COLOR_BLUE, S_COLOR_WHITE);
				CG_Printf ("%s%i - %s", S_COLOR_BLUE, i, S_COLOR_WHITE);
			}
			anim_data[rounder][counter] = i;
			rounder++;
			if (rounder > 3){
				rounder = 0;
				if (debug)
					CG_Printf ("%s anim: %i%s\n", S_COLOR_BLUE, counter, S_COLOR_WHITE);
				counter++;
				if(counter == PMODEL_MAX_ANIMATIONS)
					break;
			}
		}
	}
	
	CG_Free(buf);
	
	//it must contain at least as many animations as a Q3 script to be valid
	if(counter-1 < LEGS_TURN) {
		CG_Printf ("%sScript: Error: Not enough animations(%i) at animations script: %s%s\n", S_COLOR_YELLOW, counter, filename, S_COLOR_WHITE);
		return qfalse;
	}
	
	//animation ANIM_NONE (0) is always at frame 0, and it's never 
	//received from the game, but just used on the client when none
	//animation was ever set for a model (head).
	
	anim_data[0][ANIM_NONE] = 0;
	anim_data[1][ANIM_NONE]	= 0;
	anim_data[2][ANIM_NONE]	= 1;
	anim_data[3][ANIM_NONE] = 10;
	
	//reorganize to make my life easier
	for (i=0 ; i<counter ; i++)
	{
		pmodelinfo->firstframe[i] = anim_data[0][i];

		if (islastframe)
			pmodelinfo->lastframe[i] = anim_data[1][i];
		else
			pmodelinfo->lastframe[i] = ((anim_data[0][i]) + (anim_data[1][i]));

		pmodelinfo->loopingframes[i] = anim_data[2][i];
		
		//HACK: make weapon switching animations stuck at 10 fps
		if (i == TORSO_FLIPOUT || i == TORSO_FLIPIN || anim_data[3][i] < 10)//never allow less than 10 fps
			anim_data[3][i] = 10;
		
		pmodelinfo->frametime[i] = 1000/anim_data[3][i];
	}
	
	if (!nooffset)
	{
		//find offset (Why did they do this???)  
		offset = pmodelinfo->firstframe[LEGS_CRWALK] - pmodelinfo->firstframe[TORSO_GESTURE];
		
		//Remove offset from the lower & extra animations
		for (i = LEGS_CRWALK; i < counter; ++i){
			pmodelinfo->firstframe[i] -= offset;
			pmodelinfo->lastframe[i] -= offset;
		}
		
		if (debug)
			CG_Printf ("%sScript: adding offset on lower frames (Q3 format)%s\n", S_COLOR_BLUE, S_COLOR_WHITE);
	}
	
	//Fix the rates to fit my looping calculations
	for (i = 0; i < counter; ++i)
	{
		if (pmodelinfo->loopingframes[i])
			pmodelinfo->loopingframes[i] -= 1;
		if (pmodelinfo->lastframe[i])
			pmodelinfo->lastframe[i] -= 1;
	}
	
	//Alljumps:	 I use 3 jump animations for bunnyhoping
	//animation support. But they will only be loaded as
	//bunnyhoping animations if the keyword "alljumps" is
	//present at the animation.cfg. Otherwise, LEGS_JUMP1
	//will be used for all the jump styles.
	
	if (!alljumps)
	{
		CG_CopyAnimation (pmodelinfo, LEGS_JUMP3, LEGS_JUMP1);
		CG_CopyAnimation (pmodelinfo, LEGS_JUMP3ST, LEGS_JUMP1ST);
		CG_CopyAnimation (pmodelinfo, LEGS_JUMP2, LEGS_JUMP1);
		CG_CopyAnimation (pmodelinfo, LEGS_JUMP2ST, LEGS_JUMP1ST);
	}
	
	//HACK Uncomplete scripts (for Q3 player models)
	//----------------------------------------------------
	counter--;
	
	if (counter < TORSO_RUN)
		CG_CopyAnimation (pmodelinfo, TORSO_RUN, TORSO_STAND);
	if (counter < TORSO_DROPHOLD)
		CG_CopyAnimation (pmodelinfo, TORSO_DROPHOLD, TORSO_STAND);
	if (counter < TORSO_DROP)
		CG_CopyAnimation (pmodelinfo, TORSO_DROP, TORSO_ATTACK2);
	if (counter < TORSO_PAIN1)
		CG_CopyAnimation (pmodelinfo, TORSO_PAIN1, TORSO_STAND2);
	if (counter < TORSO_PAIN2)
		CG_CopyAnimation (pmodelinfo, TORSO_PAIN2, TORSO_STAND2);
	if (counter < TORSO_PAIN3)
		CG_CopyAnimation (pmodelinfo, TORSO_PAIN3, TORSO_STAND2);
	if (counter < TORSO_SWIM)
		CG_CopyAnimation (pmodelinfo, TORSO_SWIM, TORSO_STAND);
	
	if (counter < LEGS_WALKBACK)
		CG_CopyAnimation (pmodelinfo, LEGS_WALKBACK, LEGS_RUNBACK);
	if (counter < LEGS_WALKLEFT)
		CG_CopyAnimation (pmodelinfo, LEGS_WALKLEFT, LEGS_WALKFWD);
	if (counter < LEGS_WALKRIGHT)
		CG_CopyAnimation (pmodelinfo, LEGS_WALKRIGHT, LEGS_WALKFWD);
	if (counter < LEGS_RUNLEFT)
		CG_CopyAnimation (pmodelinfo, LEGS_RUNLEFT, LEGS_RUNFWD);
	if (counter < LEGS_RUNRIGHT)
		CG_CopyAnimation (pmodelinfo, LEGS_RUNRIGHT, LEGS_RUNFWD);
	if (counter < LEGS_SWIM)
		CG_CopyAnimation (pmodelinfo, LEGS_SWIM, LEGS_SWIMFWD);

	return qtrue;
}


/*
================
CG_LoadPModelWeaponSet

  These models are also used for the 1st person view attached weapon
================
*/
void CG_LoadPModelWeaponSet( pmodel_t *pmodel )
{
	int 		i;

	for (i = 0; i < cgs.numWeaponModels; i++) 
	{
		pmodel->weaponIndex[i] = CG_RegisterWeaponModel ( cgs.weaponModels[i] );
		
		if (!cg_vwep->integer)
			break; // only do weapon 0
	}

	//special case for weapon 0. Must always load the animation script
	if (!pmodel->weaponIndex[0])
		pmodel->weaponIndex[0] = CG_CreateWeaponZeroModel(cgs.weaponModels[0]);
}


/*
================
CG_PSkinUpdateRegistration
================
*/
qboolean CG_PSkinUpdateRegistration ( pskin_t *pSkin, char *filename )
{
	int 		i;
	char		skin_name[MAX_QPATH];
	char		path[MAX_QPATH];
	char		scratch[MAX_QPATH];

	//extract path & skin_name from filename.
	char *t = NULL;
	char *s;

	//find the word after the last slash
	s = filename;
	t = strstr ( s, "/" );
	while (t)
	{
		s = t+1;
		t = strstr ( s, "/" );
	}
	strcpy(skin_name, s);

	//now remove it from the path
	Q_snprintfz (path, strlen(filename) - strlen(skin_name), "%s", filename);

	
	//load skinfiles for each model
	for (i=0; i<PMODEL_PARTS; i++)
	{
		Q_snprintfz (scratch, sizeof(scratch), "%s/%s_%s.skin", path, pmPartNames[i], skin_name);
		pSkin->skin[i] = trap_R_RegisterSkinFile (scratch);
		
		//failed: clean up the wrong stuff
		if (!pSkin->skin[i]) {
			int	j;
			pSkin->name[0] = 0;
			for (j = 0; j < PMODEL_PARTS; j++)
				pSkin->skin[i] = NULL;
			return qfalse;
		}
	}

	//load icon
	Q_snprintfz (scratch, sizeof(scratch), "%s/icon_%s", path, skin_name);
	pSkin->icon = trap_R_RegisterPic (scratch);

	strcpy (pSkin->name, filename);
	return qtrue;
}


/*
===============
CG_FindPSkinSpot
===============
*/
struct pskin_s *CG_FindPSkinSpot ( char *filename )
{
	int 			i;
	int				freespot = -1;

	for (i=0; i<CG_MAX_PSKINS; i++)
	{
		if (cg_pSkins[i].inuse == qtrue)
		{
			if (!Q_stricmp (cg_pSkins[i].name, filename))//found it
			{
				if (cg_debugPlayerModels->integer)
					CG_Printf ("%spSkin: found at spot %i: %s\n", S_COLOR_WHITE, i, filename);

				return &cg_pSkins[i];
			}
		}
		else if (freespot < 0)
			freespot = i;
	}

	if (freespot < 0)
		CG_Error ("%sCG_FindPSkinsSpot: Couldn't find a free player skin spot%s", S_COLOR_RED, S_COLOR_WHITE);

	if (cg_debugPlayerModels->integer)
		CG_Printf ("%spSkin: assigned free spot %i for pSkin %s\n", S_COLOR_WHITE, freespot, filename);

	//we have a free spot
	return &cg_pSkins[freespot];
}


/*
===============
CG_RegisterPSkin
===============
*/
struct pskin_s *CG_RegisterPSkin ( char *filename )
{
	pskin_t	*pSkin;

	COM_StripExtension ( filename, filename );

	pSkin = CG_FindPSkinSpot (filename);

	if (pSkin->inuse == qtrue)
		return pSkin;

	pSkin->inuse = CG_PSkinUpdateRegistration ( pSkin, filename );

	if ( pSkin->inuse != qtrue )
	{
		if (cg_debugPlayerModels->integer)
			CG_Printf ("%spSkin: Registration failed: %s%s\n", S_COLOR_YELLOW, filename, S_COLOR_WHITE);

		memset (pSkin, 0, sizeof(pSkin));
		return NULL;
	}

	return pSkin;
}


/*
================
CG_PModelLoadModelPart
================
*/
qboolean CG_PModelLoadModelPart (pmodelinfo_t *pmodelinfo, char *model_name, int part)
{
	char	scratch[MAX_QPATH];

	Q_snprintfz (scratch, sizeof(scratch), "%s/%s.md3", model_name, pmPartNames[part]);
	pmodelinfo->model[part] = trap_R_RegisterModel (scratch);

	if (!pmodelinfo->model[part])
	{
		Q_snprintfz (scratch, sizeof(scratch), "%s/%s.skm", model_name, pmPartNames[part]);
		pmodelinfo->model[part] = trap_R_RegisterModel (scratch);
	}

	if (!pmodelinfo->model[part])
		return qfalse;

	return qtrue;
}

/*
================
CG_PModelUpdateRegistration
================
*/
qboolean CG_PModelUpdateRegistration ( pmodelinfo_t *pmodelinfo, char *filename )
{
	int 		i;
	qboolean 	loaded_model;
	char	anim_filename[MAX_QPATH];
	
	for (i=0; i<PMODEL_PARTS; i++)
	{
		loaded_model = CG_PModelLoadModelPart ( pmodelinfo, filename, i );
		if (!loaded_model)
			break;
	}
	
	if (loaded_model)
	{
		Q_snprintfz (anim_filename, sizeof(anim_filename), "%s/animation.cfg", filename);
		loaded_model = CG_ParseAnimationScript ( pmodelinfo, anim_filename );
	}
	
	//clean up the wrong stuff
	if (!loaded_model){
		pmodelinfo->model_name[0] = 0;
		for (i = 0; i < PMODEL_PARTS; i++)
			pmodelinfo->model[i] = NULL;
		return qfalse;
	}

	strcpy (pmodelinfo->model_name, filename);
	return qtrue;
}


/*
===============
CG_FindPModelSpot
===============
*/
struct pmodelinfo_s *CG_FindPModelSpot ( char *filename )
{
	int 			i;
	int				freespot = -1;

	for (i=0; i<CG_MAX_PMODELS; i++)
	{
		if (cg_PModelInfos[i].inuse == qtrue)
		{
			if (!Q_stricmp (cg_PModelInfos[i].model_name, filename))//found it
			{
				if (cg_debugPlayerModels->integer)
					CG_Printf ("%sPModel: found at spot %i: %s\n", S_COLOR_WHITE, i, filename);

				return &cg_PModelInfos[i];
			}
		}
		else if (freespot < 0)
			freespot = i;
	}

	if (freespot < 0)
		CG_Error ("%sCG_FindPModelSpot: Couldn't find a free player model spot%s", S_COLOR_RED, S_COLOR_WHITE);

	if (cg_debugPlayerModels->integer)
		CG_Printf ("%sPModel: assigned free spot %i for pmodel %s\n", S_COLOR_WHITE, freespot, filename);

	//we have a free spot
	return &cg_PModelInfos[freespot];
}


/*
===============
CG_RegisterPModel

  PModelinfo is not exactly the model, but the indexes of the
  models contained in the pmodel and it's animation data
===============
*/
struct pmodelinfo_s *CG_RegisterPModel ( char *filename )
{
	pmodelinfo_t	*pmodelinfo;

	COM_StripExtension ( filename, filename );

	pmodelinfo = CG_FindPModelSpot (filename);

	if (pmodelinfo->inuse == qtrue)
		return pmodelinfo;

	pmodelinfo->inuse = CG_PModelUpdateRegistration ( pmodelinfo, filename );

	if ( pmodelinfo->inuse != qtrue )
	{
		if (cg_debugPlayerModels->integer)
			CG_Printf ("%sPModel: Registration failed: %s%s\n", S_COLOR_YELLOW, filename, S_COLOR_WHITE);

		memset (pmodelinfo, 0, sizeof(pmodelinfo));
		return NULL;
	}

	return pmodelinfo;
}


/*
================
CG_RegisterBasePModel

  Default fallback replacements
================
*/
void CG_RegisterBasePModel(void)
{
	char	filename[MAX_QPATH];

	//pmodelinfo
	Q_snprintfz (filename, sizeof(filename), "%s/%s", "models/players", DEFAULT_PLAYERMODEL);
	cgs.basePModelInfo = CG_RegisterPModel ( filename );
	//pskin
	Q_snprintfz (filename, sizeof(filename), "%s/%s/%s","models/players" , DEFAULT_PLAYERMODEL, DEFAULT_PLAYERSKIN);
	cgs.basePSkin = CG_RegisterPSkin (filename);

	//Default player model must load perfect
	if(!cgs.basePModelInfo)
		CG_Error ( "%s'Default Player Model'(%s): failed to load%s\n", S_COLOR_RED, DEFAULT_PLAYERMODEL, S_COLOR_WHITE );

	if(!cgs.basePSkin)
		CG_Error ( "%s'Default Player Model'(%s): Skin (%s) failed to load%s\n", S_COLOR_RED, DEFAULT_PLAYERMODEL, DEFAULT_PLAYERMODEL, S_COLOR_WHITE );

	if (!cgs.basePSkin->icon)
		CG_Error ( "%s'Default Player Model'(%s): Icon failed to load%s\n", S_COLOR_RED, DEFAULT_PLAYERMODEL, S_COLOR_WHITE );
}


/*
================
CG_LoadClientPmodel
Register a special client pmodel (not indexed)
================
*/
void CG_LoadClientPmodel ( int centNum, char *model_name, char *skin_name )
{
	pmodel_t			*pmodel;
	char				filename[MAX_QPATH];
	char				*path = "models/players";
	char				skin_filename[MAX_QPATH];
	

	if (centNum < 0)
		centNum = 0;
	
	pmodel = &cg_entPModels[centNum];

	//load user defined pmodelinfo
	Q_snprintfz (filename, sizeof(filename), "%s/%s",path, model_name);
	pmodel->pmodelinfo = CG_RegisterPModel( filename );
	//load skin
	Q_snprintfz (skin_filename, sizeof(skin_filename), "%s/%s/%s",path, model_name, skin_name);
	pmodel->pSkin = CG_RegisterPSkin (skin_filename);
	
	
	//didn't work: try DEFAULT_PLAYERMODEL with the user defined skin (for ctf skins)
	if ((!pmodel->pmodelinfo || !pmodel->pSkin) && Q_stricmp(model_name, DEFAULT_PLAYERMODEL))
	{
		strcpy(model_name, DEFAULT_PLAYERMODEL);
		Q_snprintfz (filename, sizeof(filename), "%s/%s",path, model_name);
		pmodel->pmodelinfo = CG_RegisterPModel( filename );
		//load skin
		Q_snprintfz (skin_filename, sizeof(skin_filename), "%s/%s/%s",path, model_name, skin_name);
		pmodel->pSkin = CG_RegisterPSkin (skin_filename);
	}
	
	//if still didn't load, it means that the DEFAULT_PLAYERMODEL
	//didn't have the custom skin so back to DEFAULT_PLAYERSKIN
	if (!pmodel->pSkin)
	{
		strcpy(skin_name, DEFAULT_PLAYERSKIN);
		Q_snprintfz (skin_filename, sizeof(skin_filename), "%s/%s/%s",path, model_name, skin_name);
		pmodel->pSkin = CG_RegisterPSkin (skin_filename);
	}

	if (!pmodel->pmodelinfo)
		CG_Error ( "Couldn't load default player model" );

	//store weaponIndexes
	CG_LoadPModelWeaponSet ( pmodel );
	

	//update the frames for the new animation script
	if (!pmodel->anim.current[UPPER] || !pmodel->anim.current[LOWER] || pmodel->anim.current[HEAD] != ANIM_NONE )
	{
		pmodel->anim.current[UPPER] = TORSO_STAND;
		pmodel->anim.current[LOWER] = LEGS_STAND;
		pmodel->anim.current[HEAD] = ANIM_NONE;
	}
	
	pmodel->anim.oldframe[UPPER] = pmodel->anim.frame[UPPER] = pmodel->pmodelinfo->lastframe[pmodel->anim.current[UPPER]];
	pmodel->anim.oldframe[LOWER] = pmodel->anim.frame[LOWER] = pmodel->pmodelinfo->lastframe[pmodel->anim.current[LOWER]];
	pmodel->anim.oldframe[HEAD] = pmodel->anim.frame[HEAD] = pmodel->pmodelinfo->lastframe[pmodel->anim.current[HEAD]];
}



//======================================================================
//							tools
//======================================================================


/*
===============
CG_GrabTag
===============
*/

qboolean CG_GrabTag(orientation_t *tag, entity_t *ent, char *tagname)
{
	if (!ent->model)
		return qfalse;

	return trap_R_LerpTag( tag, ent->model, ent->frame, ent->oldframe, ent->backlerp, tagname );
}


/*
===============
CG_PlaceRotatedModelOnTag
===============
*/
void CG_PlaceRotatedModelOnTag (entity_t *ent, entity_t *dest, orientation_t *tag)
{
	int		i;
	vec3_t	tmpAxis[3];

	VectorCopy( dest->origin, ent->origin );
	VectorCopy (dest->lightingOrigin, ent->lightingOrigin);
	
	for ( i = 0 ; i < 3 ; i++ )
		VectorMA( ent->origin, tag->origin[i], dest->axis[i], ent->origin );

	VectorCopy (ent->origin, ent->oldorigin);

	Matrix_Multiply( ent->axis, tag->axis, tmpAxis );
	Matrix_Multiply( tmpAxis, dest->axis, ent->axis );
}


/*
===============
CG_PlaceModelOnTag
===============
*/
void CG_PlaceModelOnTag (entity_t *ent, entity_t *dest, orientation_t *tag)
{
	int i;
	
	VectorCopy (dest->origin, ent->origin);
	VectorCopy (dest->lightingOrigin, ent->lightingOrigin);
	
	for ( i = 0 ; i < 3 ; i++ )
		VectorMA( ent->origin, tag->origin[i], dest->axis[i], ent->origin );
	
	VectorCopy (ent->origin, ent->oldorigin);
	
	Matrix_Multiply( tag->axis, dest->axis, ent->axis );
}

/*
===============
CG_MoveToTag
===============
*/
void CG_MoveToTag ( orientation_t *result, orientation_t *ref, orientation_t *tag)
{
	int i;
	
	VectorCopy (ref->origin, result->origin);
	
	for ( i = 0 ; i < 3 ; i++ )
		VectorMA( result->origin, tag->origin[i], ref->axis[i], result->origin );
	
	Matrix_Multiply( tag->axis, ref->axis, result->axis );
}

/*
===============
CG_PModel_CalcProjectionSource
===============
*/
qboolean CG_PModel_CalcProjectionSource( pmodel_t *pmodel, orientation_t *projection )
{
	orientation_t tag_weapon;
	orientation_t	result1, ref;

	if(!pmodel->pmodelinfo)
		return qfalse;

	if( !pmodel->ents[UPPER].model || !pmodel->ents[LOWER].model )
		return qfalse;

	//the code asumes that the pmodel ents are up-to-date
	if(!trap_R_LerpTag( &tag_weapon, pmodel->ents[UPPER].model, pmodel->ents[UPPER].frame, pmodel->ents[UPPER].oldframe, pmodel->ents[UPPER].backlerp, "tag_weapon" ) )
		return qfalse;

	VectorCopy( pmodel->ents[UPPER].origin, ref.origin );
	Matrix_Copy( pmodel->ents[UPPER].axis, ref.axis );
	CG_MoveToTag ( &result1, &ref, &tag_weapon );

	//our origin is now at tag_weapon. Move it to tag_projectionsource
	CG_MoveToTag ( projection, &result1, &pmodel->pweapon.weaponInfo->tag_projectionsource );
	return qtrue;
}

/*
===============
CG_PModel_CalcBrassSource
===============
*/
void CG_PModel_CalcBrassSource( pmodel_t *pmodel, orientation_t *projection )
{
	orientation_t	tag_weapon;
	orientation_t	ref;

	//the code asumes that the pmodel ents are up-to-date
	if(!pmodel->pmodelinfo || !pmodel->ents[UPPER].model || !pmodel->ents[LOWER].model 
		|| !trap_R_LerpTag( &tag_weapon, pmodel->ents[UPPER].model, pmodel->ents[UPPER].frame, pmodel->ents[UPPER].oldframe, pmodel->ents[UPPER].backlerp, "tag_weapon" ) )
	{
		VectorCopy(pmodel->ents[LOWER].origin, projection->origin);
		projection->origin[2] += 16;
		Matrix_Identity(projection->axis);
	}
	//brass projection is simply tag_weapon
	VectorCopy( pmodel->ents[UPPER].origin, ref.origin );
	Matrix_Copy( pmodel->ents[UPPER].axis, ref.axis );
	CG_MoveToTag ( projection, &ref, &tag_weapon );
}

/*
===============
CG_AddShellEffects
===============
*/
void CG_AddQuadShell(entity_t *ent)
{
	entity_t	shell;

	shell = *ent;
	shell.customSkin = NULL;

	if ( shell.flags & RF_WEAPONMODEL )
		shell.customShader = CG_MediaShader ( cgs.media.shaderQuadWeapon );
	else
		shell.customShader = CG_MediaShader ( cgs.media.shaderPowerupQuad );

	trap_R_AddEntityToScene( &shell );
}


void CG_AddPentShell (entity_t *ent)
{
	entity_t	shell;

	shell = *ent;
	shell.customSkin = NULL;

	if ( shell.flags & RF_WEAPONMODEL )
		shell.customShader = CG_MediaShader ( cgs.media.shaderPowerupPenta );//fixme: look for a PENTA shader for view weapon
	else
		shell.customShader = CG_MediaShader ( cgs.media.shaderPowerupPenta );

	trap_R_AddEntityToScene( &shell );
}


void CG_AddShellEffects (entity_t *ent, int effects)
{
	// quad and pent can do different things on client

	if (effects & EF_QUAD) 
		CG_AddQuadShell (ent);
/*
	if (effects & EF_PENT) 
		CG_AddQuadShell (ent); //jalfixme
		//CG_AddPentShell (ent);*/
}

/*
===============
CG_AddColorShell
===============
*/
void CG_AddColorShell_Do (entity_t *ent, int renderfx)
{
	entity_t	shell;
	int			i;
	vec4_t shadelight = { 0, 0, 0, 0.3 };

	memset (&shell, 0, sizeof(shell));
	
	shell = *ent;
	shell.customSkin = NULL;
	
	if ( renderfx & RF_SHELL_RED )
		shadelight[0] = 1.0;
	if ( renderfx & RF_SHELL_GREEN )
		shadelight[1] = 1.0;
	if ( renderfx & RF_SHELL_BLUE )
		shadelight[2] = 1.0;
	
	for (i=0 ; i<4 ; i++)
		shell.color[i] = shadelight[i] * 255;
	
	if ( ent->flags & RF_WEAPONMODEL )
		return;//fixme: try the shell shader for viewweapon, or build a good one
	else
		shell.customShader = CG_MediaShader ( cgs.media.shaderShellEffect );

	trap_R_AddEntityToScene( &shell );
}

void CG_AddColorShell (entity_t *ent, int renderfx)
{
	if ( !(renderfx & (RF_SHELL_GREEN | RF_SHELL_RED | RF_SHELL_BLUE)) )
		return;

	CG_AddColorShell_Do (ent, renderfx);
}


//======================================================================
//							animations
//======================================================================


/*
===============
CG_PModelAnimToFrame

  Transforms the values inside state->frame into 3 animation values
  for head, torso and legs (head is always zero right now). It also
  uses those values to handle the animation frames clientside, as 
  also does the backlerp calculation for variable framerates.
  
  The frames and other data is stored inside animationinfo_t. Actually
  only clientinfo has structs of this type inside, but they could also
  be applied to a client-side monsterinfo handler.
  
  It works like this:
  -for animations reveived by BASIC_CHANNEL:
  When a new animation is recieved, the new animation is launched.
  While 0 is recieved, it advances the frames in last recieved animation.
  When the frame gets to the end of the animation, it looks for the
  looping value and sends the frame back to (lastframe - loopingframes)
  and so on. If the looping value is 0, it holds the animation in the
  last frame until a new animation is recieved.

  -for animations reveived by EVENT_CHANNEL:
  The playing BASIC_CHANNEL animation is put back on the animation
  buffer, the EVENT_CHANNEL animation is played, and when it comes to 
  it's end,the BASIC_CHANNEL animation at buffer (the same as we put in, 
  or a new one received by that time) is promoted to be played.
===============
*/
void CG_PModelAnimToFrame (pmodelinfo_t *pmodelinfo, animationinfo_t *anim)
{
	int i;
	
	for (i=0; i<PMODEL_PARTS; i++)
	{
		//interpolate
		if (cg.time < anim->nextframetime[i])
		{
			anim->backlerp[i] = 1.0f - ((cg.time - anim->prevframetime[i])/(anim->nextframetime[i] - anim->prevframetime[i]));
			if (anim->backlerp[i] > 1)
				anim->backlerp[i] = 1;
			else if (anim->backlerp[i] < 0)
				anim->backlerp[i] = 0;
			
			continue;
		}
		
		//advance frames
		anim->oldframe[i] = anim->frame[i];
		anim->frame[i]++;
		
		//looping
		if (anim->frame[i] > pmodelinfo->lastframe[anim->current[i]])
		{
			if (anim->currentChannel[i])	
				anim->currentChannel[i] = BASIC_CHANNEL;//kill

			anim->frame[i] = (pmodelinfo->lastframe[anim->current[i]] - pmodelinfo->loopingframes[anim->current[i]]);
		}

		//NEWANIM
		if ( anim->buffer[EVENT_CHANNEL].newanim[i] )
		{
			//backup if basic
			if (anim->buffer[BASIC_CHANNEL].newanim[i] +
				anim->currentChannel[i] == BASIC_CHANNEL){//both are zero
				anim->buffer[BASIC_CHANNEL].newanim[i] = anim->current[i];
			}
			//set up
			anim->current[i] = anim->buffer[EVENT_CHANNEL].newanim[i];
			anim->frame[i] = pmodelinfo->firstframe[anim->current[i]];
			anim->currentChannel[i] = EVENT_CHANNEL;
			anim->buffer[EVENT_CHANNEL].newanim[i] = 0;
		}
		else if ( anim->buffer[BASIC_CHANNEL].newanim[i]
			&& ( anim->currentChannel[i] != EVENT_CHANNEL ) )
		{
			//set up
			anim->current[i] = anim->buffer[BASIC_CHANNEL].newanim[i];
			anim->frame[i] = pmodelinfo->firstframe[anim->current[i]];
			anim->currentChannel[i] = BASIC_CHANNEL;
			anim->buffer[BASIC_CHANNEL].newanim[i] = 0;
		}

		//updated frametime
		anim->prevframetime[i] = cg.time;
		anim->nextframetime[i] = cg.time + pmodelinfo->frametime[anim->current[i]];
		anim->backlerp[i] = 1.0f;
	}
}


/*
===============
CG_ClearEventAnimations

  Called from CG_PModelUpdateState if the new state has a EV_TELEPORT.
  This is done so respawning can reset animations, because event
  animations could persist after dying
===============
*/
void CG_ClearEventAnimations(entity_state_t *state)
{
	int i;
	pmodel_t	*pmodel;

	pmodel = &cg_entPModels[state->number];

	for ( i = 0 ; i < PMODEL_PARTS ; i++ )
	{
		//clean up the buffer
		pmodel->anim.buffer[EVENT_CHANNEL].newanim[i] = 0;

		//finish up currents
		if (pmodel->anim.currentChannel[i] == EVENT_CHANNEL)
		{
			pmodel->anim.frame[i] = pmodel->pmodelinfo->lastframe[pmodel->anim.current[i]];
			pmodel->anim.currentChannel[i] = BASIC_CHANNEL;
		}
	}
}


/*
===============
CG_AddPModelAnimation
===============
*/
void CG_AddPModelAnimation ( pmodel_t *pmodel, int loweranim, int upperanim, int headanim, int channel)
{
	int	i;
	int newanim[PMODEL_PARTS];
	animationbuffer_t	*buffer;

	newanim[LOWER] = loweranim;
	newanim[UPPER] = upperanim;
	newanim[HEAD] = headanim;

	buffer = &pmodel->anim.buffer[channel];

	for ( i = 0 ; i < PMODEL_PARTS ; i++ )
	{
		//ignore new events if in death
		if (channel && buffer->newanim[i] && (buffer->newanim[i] < TORSO_GESTURE))
			continue;

		if (newanim[i] && (newanim[i] < PMODEL_MAX_ANIMS))
		{
			buffer->newanim[i] = newanim[i];
		}
	}
}

/*
===============
CG_AddAnimationFromState
===============
*/
void CG_AddAnimationFromState ( entity_state_t *state, int loweranim, int upperanim, int headanim, int channel)
{
	CG_AddPModelAnimation ( &cg_entPModels[state->number], loweranim, upperanim, headanim, channel);
}

/*
==============
CG_PModels_AddFireEffects

  launch fire animations from events
==============
*/
void CG_PModels_AddFireEffects ( entity_state_t *state )
{
	pmodel_t *pmodel;

	if ( state->effects & EF_CORPSE )
		return;

	//if it's the user in 1st person, viewapon code handles this
	if( state->number == cg.chasedNum+1 && !cg.thirdPerson )
		return;

	pmodel = &cg_entPModels[state->number];

	//activate weapon flash
	if (cg_weaponFlashes->integer)
		pmodel->pweapon.flashtime = cg.time + (int)pmodel->pmodelinfo->frametime[TORSO_ATTACK1];
		//pmodel->pweapon.flashtime = cg.time + (int)((pmodel->pmodelinfo->frametime[TORSO_ATTACK1]/4)*3);


	//add effects
	switch ( state->weapon )
	{

	case WEAP_GUNBLADE:
		CG_AddPModelAnimation ( pmodel, 0, TORSO_ATTACK1, 0, EVENT_CHANNEL);
		break;
	case WEAP_SHOCKWAVE:
		CG_AddPModelAnimation ( pmodel, 0, TORSO_ATTACK1, 0, EVENT_CHANNEL);
		break;
	case WEAP_ROCKETLAUNCHER:
		CG_AddPModelAnimation ( pmodel, 0, TORSO_ATTACK1, 0, EVENT_CHANNEL);
		break;
	case WEAP_GRENADELAUNCHER:
		CG_AddPModelAnimation ( pmodel, 0, TORSO_ATTACK1, 0, EVENT_CHANNEL);
		break;
	case WEAP_PLASMAGUN:
		CG_AddPModelAnimation ( pmodel, 0, TORSO_ATTACK1, 0, EVENT_CHANNEL);
		break;
	case WEAP_RIOTGUN:
		CG_AddPModelAnimation ( pmodel, 0, TORSO_ATTACK1, 0, EVENT_CHANNEL);
		break;
	case WEAP_ELECTROBOLT:
		CG_AddPModelAnimation ( pmodel, 0, TORSO_ATTACK1, 0, EVENT_CHANNEL);
		break;
	
	default:
		break;
	}
}

//======================================================================
//							player model
//======================================================================


/*
==================
CG_PModelUpdateState

  Called each new serverframe
==================
*/
void CG_PModelUpdateState (entity_state_t *state)
{
	int newanim[PMODEL_PARTS];
	entity_state_t *oldstate;
	int			i;
	pmodel_t	*pmodel;
	
	pmodel = &cg_entPModels[state->number];
	
	if (state->modelindex != 255)	 
	{
		//in ctf, corpses mirror it's client pmodel (ctf skins)
		if( (cgs.gametype == GAMETYPE_CTF || cgs.gametype == GAMETYPE_TDM ||
			cgs.gametype == GAMETYPE_DUEL ) && state->effects & EF_CORPSE ) {
			pmodel->pmodelinfo = cg_entPModels[(state->skinnum & 0xff)+1].pmodelinfo;
			pmodel->pSkin = cg_entPModels[(state->skinnum & 0xff)+1].pSkin;
		} else {	//indexed
			pmodel->pmodelinfo = cgs.pModelsIndex[state->modelindex];
			pmodel->pSkin = cgs.pSkinsIndex[state->skinnum];
		}
	}

	//fallback
	if (!pmodel->pmodelinfo || !pmodel->pSkin){
		pmodel->pmodelinfo = cgs.basePModelInfo;
		pmodel->pSkin = cgs.basePSkin;
	}
	
	//update ents models & skins
	for ( i = LOWER ; i < PMODEL_PARTS ; i++ )
	{
		pmodel->ents[i].scale = 1.0f;
		pmodel->ents[i].rtype = RT_MODEL;
		pmodel->ents[i].customShader = NULL;
		pmodel->ents[i].model = pmodel->pmodelinfo->model[i];
		pmodel->ents[i].customSkin = pmodel->pSkin->skin[i];
	}
	
	//Update pweapon
	pmodel->pweapon.weaponInfo = CG_GetWeaponFromPModelIndex( pmodel, state->weapon );

	//update parts rotation angles
	for ( i = LOWER ; i < PMODEL_PARTS ; i++ )
		VectorCopy(pmodel->angles[i], pmodel->oldangles[i]);

	if(!(state->effects & EF_CORPSE))
	{
		//lower has horizontal direction, and zeroes vertical
		pmodel->angles[LOWER][PITCH] = 0;
		pmodel->angles[LOWER][YAW] = state->angles[YAW];
		pmodel->angles[LOWER][ROLL] = 0;
		
		//upper marks vertical direction (total angle, so it fits aim)
		if (state->angles[PITCH] > 180)
			pmodel->angles[UPPER][PITCH] = (-360 + state->angles[PITCH]);
		else
			pmodel->angles[UPPER][PITCH] = state->angles[PITCH];
		
		pmodel->angles[UPPER][YAW] = 0;
		pmodel->angles[UPPER][ROLL] = 0;
		
		//head adds a fraction of vertical angle again
		if (state->angles[PITCH] > 180)
			pmodel->angles[HEAD][PITCH] = (-360 + state->angles[PITCH])/3;
		else
			pmodel->angles[HEAD][PITCH] = state->angles[PITCH]/3;
		
		pmodel->angles[HEAD][YAW] = 0;
		pmodel->angles[HEAD][ROLL] = 0;
	}

	//Spawning (EV_TELEPORT) forces nobacklerp and the interruption of EVENT_CHANNEL animations
	if ((state->events[0] == EV_TELEPORT) || (state->events[1] == EV_TELEPORT))
	{
		CG_ClearEventAnimations(state);
		CG_AddAnimationFromState ( state, (state->frame)&0x3F, (state->frame>>6)&0x3F, (state->frame>>12)&0xF, BASIC_CHANNEL);//(jalPVSfix2)
		//set up old frame
		for ( i = LOWER ; i < PMODEL_PARTS ; i++ )
			pmodel->anim.nextframetime[i] = cg.time;

		CG_PModelAnimToFrame( pmodel->pmodelinfo, &pmodel->anim );
		pmodel->anim.oldframe[LOWER] = pmodel->anim.frame[LOWER];
		pmodel->anim.oldframe[UPPER] = pmodel->anim.frame[UPPER];
		pmodel->anim.oldframe[HEAD] = pmodel->anim.frame[HEAD];

		for ( i = LOWER ; i < PMODEL_PARTS ; i++ )
			VectorCopy(pmodel->angles[i], pmodel->oldangles[i]);

		return;
	}

	oldstate = &cg_entities[state->number].prev;

	//filter repeated animations coming from state->frame
	newanim[LOWER] = (state->frame&0x3F) * ((state->frame &0x3F) != (oldstate->frame &0x3F));
	newanim[UPPER] = (state->frame>>6 &0x3F) * ((state->frame>>6 &0x3F) != (oldstate->frame>>6 &0x3F));
	newanim[HEAD] = (state->frame>>12 &0xF) * ((state->frame>>12 &0xF) != (oldstate->frame>>12 &0xF));
	
	CG_AddAnimationFromState ( state, newanim[LOWER], newanim[UPPER], newanim[HEAD], BASIC_CHANNEL);
}


/*
==================
CG_PModelsUpdateStates

  Called each new serverframe
==================
*/
void CG_PModelsUpdateStates (void)
{
	int					pnum;
	entity_state_t		*state;

	for ( pnum = 0; pnum < cg.frame.numEntities; pnum++ )
	{
		state = &cg_parseEntities[(cg.frame.parseEntities+pnum)&(MAX_PARSE_ENTITIES-1)];
		if ( (state->type & ~ET_INVERSE) == ET_PLAYER ) {
			CG_PModelUpdateState (state);
		}
	}
}

void CG_AddOutLineEffect(entity_t *ent, int effects)
{
	entity_t	shell;

	if( !(effects & EF_OUTLINE) )
		return;

	//if( !cg_PModelOutlines->integer )
	//	return;

	shell = *ent;
	shell.customSkin = NULL;

	if ( shell.flags & RF_WEAPONMODEL )
		shell.customShader = CG_MediaShader ( cgs.media.shaderBlackModelOutline_thin );
	else
		shell.customShader = CG_MediaShader ( cgs.media.shaderBlackModelOutline );

	trap_R_AddEntityToScene( &shell );
}

void CG_AddOutLineEffect2(entity_t *ent, int effects)
{
	entity_t	shell;

	//if( !cg_PModelOutlines->integer )
	//	return;

	shell = *ent;
	shell.customSkin = NULL;
	if (effects & EF_QUAD)
		shell.customShader = trap_R_RegisterPic( "celloutline/firequad" );
/*	else if (effects & EF_PENT)
		shell.customShader = trap_R_RegisterPic( "celloutline/firequad" );*/
	else
		shell.customShader = CG_MediaShader ( cgs.media.shaderBlackModelOutline );

	trap_R_AddEntityToScene( &shell );
}

void CG_PModel_AddBusyIcon( entity_t *ent, int effects )
{
	entity_t	balloon;

	if( !(effects & EF_BUSYICON) )
		return;

	memset( &balloon, 0, sizeof( entity_t) );
	balloon.rtype = RT_SPRITE;
	balloon.model = NULL;
	balloon.flags = RF_NOSHADOW;
	balloon.radius = 12;
	balloon.customShader = trap_R_RegisterPic( PATH_BALLONCHAT_I );
	balloon.scale = 1.0f;

	balloon.origin[0] = ent->origin[0];
	balloon.origin[1] = ent->origin[1];
	balloon.origin[2] = ent->origin[2] + playerbox_stand_maxs[2] + balloon.radius + 2;
	VectorCopy(balloon.origin, balloon.oldorigin);
	Matrix_Identity( balloon.axis );

	trap_R_AddEntityToScene( &balloon );
}

/*
===============
CG_AddPModelEnt

===============
*/

//jal temp test
//#include "cg_jaltools.c"
//void CG_AddTestBox ( vec3_t origin, vec3_t mins, vec3_t maxs );
//jal temp test end

void CG_AddPModel (entity_t *ent, entity_state_t *state)
{
	orientation_t		tag;
	int					i, prev, j;
	pmodel_t			*pmodel;
	vec3_t				tmpangles;
	

	pmodel = &cg_entPModels[state->number];

	//transform animation values into frames
	CG_PModelAnimToFrame( pmodel->pmodelinfo, &pmodel->anim );
	
	for ( i = LOWER ; i < PMODEL_PARTS ; i++ )
	{
		pmodel->ents[i].backlerp = pmodel->anim.backlerp[i];
		pmodel->ents[i].frame = pmodel->anim.frame[i];
		pmodel->ents[i].oldframe = pmodel->anim.oldframe[i];
	}

	if( state->effects & EF_CORPSE ) {
		//zero rotations
		AnglesToAxis ( state->angles, pmodel->ents[LOWER].axis );
		for ( i = UPPER ; i < PMODEL_PARTS ; i++ )
			Matrix_Copy( axis_identity, pmodel->ents[i].axis );
		
	} else {
		
		for ( i = LOWER ; i < PMODEL_PARTS ; i++ ){
			//lerp rotated angles
			for ( j = 0; j < 3; j++ )
				tmpangles[j] = LerpAngle ( pmodel->oldangles[i][j], pmodel->angles[i][j], cg.lerpfrac );
			
			AnglesToAxis ( tmpangles, pmodel->ents[i].axis );
		}
	}

	// Add playermodel ents

	//LOWER needs special handling because of not being tagged
	VectorCopy (ent->origin, pmodel->ents[LOWER].origin);
	VectorCopy (ent->oldorigin, pmodel->ents[LOWER].oldorigin);
	VectorCopy (ent->lightingOrigin, pmodel->ents[LOWER].lightingOrigin);

	for ( i = LOWER ; i < PMODEL_PARTS ; i++ )
	{
		prev = i - 1;
		pmodel->ents[i].flags = ent->flags;

		//place on their respective tags
		if (i && (pmodel->ents[prev].model))	//don't try moving the first model to tag
		{
			if (CG_GrabTag( &tag, &pmodel->ents[prev], pmTagNames[prev] ))
				CG_PlaceRotatedModelOnTag(&pmodel->ents[i], &pmodel->ents[prev], &tag);
		}

		trap_R_AddEntityToScene( &pmodel->ents[i] );

		if (!pmodel->ents[i].model)	//don't continue without a model
			break;

		
		CG_AddOutLineEffect2 (&pmodel->ents[i], state->effects);
		CG_AddShellEffects (&pmodel->ents[i], state->effects);
		CG_AddColorShell (&pmodel->ents[i], state->renderfx);
	}

	//jal temp test
	//if(!(state->effects & EF_CORPSE)) {
	//	CG_AddTestBox ( pmodel->ents[LOWER].origin, playerbox_stand_mins, playerbox_stand_maxs );
	//}

	// add chat icon
	CG_PModel_AddBusyIcon( &pmodel->ents[LOWER], state->effects );

	// Add weapon model
	if (state->modelindex2 != 255 || !state->weapon) //player doesn't have a weapon
		return;

	if( CG_GrabTag( &tag, &pmodel->ents[UPPER], "tag_weapon" ) )
		CG_AddWeaponOnTag ( &pmodel->ents[UPPER], &tag, &pmodel->pweapon, state->effects | EF_OUTLINE);
}
