/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// g_combat.c

#include "g_local.h"
#include "g_gametypes.h"

/*
============
CanDamage

Returns true if the inflictor can directly damage the target.  Used for
explosions and melee attacks.
============
*/
qboolean CanDamage (edict_t *targ, edict_t *inflictor)
{
	vec3_t	dest;
	trace_t	trace;

// bmodels need special checking because their origin is 0,0,0
	if (targ->movetype == MOVETYPE_PUSH)
	{
		VectorAdd (targ->r.absmin, targ->r.absmax, dest);
		VectorScale (dest, 0.5, dest);
		trap_Trace (&trace, inflictor->s.origin, vec3_origin, vec3_origin, dest, inflictor, MASK_SOLID);
		if (trace.fraction == 1.0)
			return qtrue;
		if (&game.edicts[trace.ent] == targ)
			return qtrue;
		return qfalse;
	}
	
	trap_Trace (&trace, inflictor->s.origin, vec3_origin, vec3_origin, targ->s.origin, inflictor, MASK_SOLID);
	if (trace.fraction == 1.0)
		return qtrue;

	VectorCopy (targ->s.origin, dest);
	dest[0] += 15.0;
	dest[1] += 15.0;
	trap_Trace (&trace, inflictor->s.origin, vec3_origin, vec3_origin, dest, inflictor, MASK_SOLID);
	if (trace.fraction == 1.0)
		return qtrue;

	VectorCopy (targ->s.origin, dest);
	dest[0] += 15.0;
	dest[1] -= 15.0;
	trap_Trace (&trace, inflictor->s.origin, vec3_origin, vec3_origin, dest, inflictor, MASK_SOLID);
	if (trace.fraction == 1.0)
		return qtrue;

	VectorCopy (targ->s.origin, dest);
	dest[0] -= 15.0;
	dest[1] += 15.0;
	trap_Trace (&trace, inflictor->s.origin, vec3_origin, vec3_origin, dest, inflictor, MASK_SOLID);
	if (trace.fraction == 1.0)
		return qtrue;

	VectorCopy (targ->s.origin, dest);
	dest[0] -= 15.0;
	dest[1] -= 15.0;
	trap_Trace (&trace, inflictor->s.origin, vec3_origin, vec3_origin, dest, inflictor, MASK_SOLID);
	if (trace.fraction == 1.0)
		return qtrue;

	return qfalse;
}


/*
============
Killed
============
*/
void Killed (edict_t *targ, edict_t *inflictor, edict_t *attacker, int damage, vec3_t point, int mod)
{
	if (targ->health < -999)
		targ->health = -999;

	targ->enemy = attacker;

	//newgametypes[start]
	if( G_Gametype_Killed(targ, inflictor, attacker, damage, point, mod) )
		return;
	//newgametypes[end]

	if ((targ->r.svflags & SVF_MONSTER) && (targ->deadflag != DEAD_DEAD))
	{
//		targ->r.svflags |= SVF_CORPSE;		// now treat as a different content type
	}

	if (targ->movetype == MOVETYPE_PUSH || targ->movetype == MOVETYPE_STOP || targ->movetype == MOVETYPE_NONE)
	{	// doors, triggers, etc
		targ->die (targ, inflictor, attacker, damage, point);
		return;
	}

	targ->die (targ, inflictor, attacker, damage, point);
}


/*
================
SpawnDamage
================
*/
void SpawnDamage (int type, vec3_t origin, vec3_t normal, int damage)
{
	edict_t *event;
	event = G_SpawnEvent ( type, DirToByte (normal), origin );
	event->r.svflags = SVF_NOOLDORIGIN;
}

/*
============
T_Damage

targ		entity that is being damaged
inflictor	entity that is causing the damage
attacker	entity that caused the inflictor to damage targ
	example: targ=monster, inflictor=rocket, attacker=player

dir			direction of the attack
point		point at which the damage is being inflicted
normal		normal vector from that point
damage		amount of damage being inflicted
knockback	force to be applied against targ as a result of the damage

dflags		these flags are used to control how T_Damage works
	DAMAGE_RADIUS			damage was indirect (from a nearby explosion)
	DAMAGE_NO_ARMOR			armor does not protect from this damage
	DAMAGE_ENERGY			damage is from an energy based weapon
	DAMAGE_NO_KNOCKBACK		do not affect velocity, just view angles
	DAMAGE_BULLET			damage is from a bullet (used for ricochets)
	DAMAGE_NO_PROTECTION	kills godmode, armor, everything
============
*/
static int CheckArmor (edict_t *ent, vec3_t point, vec3_t normal, int damage, int te_sparks, int dflags)
{
	gclient_t	*client;
	int			save;
	int			armortag;
	gitem_t		*armor;

	if (!damage)
		return 0;

	client = ent->r.client;

	if (!client)
		return 0;

	if (dflags & DAMAGE_NO_ARMOR)
		return 0;

	armortag = ArmorTag (ent->r.client->pers.inventory);
	if (!armortag)
		return 0;

	armor = game.items[armortag];

	//jal: make the armors save the same damage to every type of weapon
	//if (dflags & DAMAGE_ENERGY)
	//	save = ceil(((gitem_armor_t *)armor->info)->energy_protection*damage);
	//else
	save = ceil(((gitem_armor_t *)armor->info)->protection*damage);
	if (save >= client->pers.inventory[armortag])
		save = client->pers.inventory[armortag];

	if (!save)
		return 0;

	client->pers.inventory[armortag] -= save;
	
	//SpawnDamage (te_sparks, point, normal, save); //this spawned the armor particles

	return save;
}

/*
================
G_AddDamageForFrame - moveme to g_gameteams?
================
*/
qboolean G_IsTeamDamage (edict_t *targ, edict_t *attacker)
{
	if( !G_Gametype_IsTeamBased() ) 
		return qfalse;
	
	if( targ->gameTeam && attacker->gameTeam &&
		targ->gameTeam == attacker->gameTeam &&
		targ != attacker )
		return qtrue;
	
	return qfalse;
}


/*
================
G_AddDamageForFrame
================
*/
void G_BlendFrameDamage( edict_t *ent, int damage, int *old_damage, vec3_t point, vec3_t dir, vec3_t old_point, vec3_t old_dir )
{
	vec3_t	offset;
	float	frac;
	int		i;

	if( !dir ) {

	}

	if( !point )
		VectorSet( offset, 0, 0, ent->viewheight );
	else
		VectorSubtract( point, ent->s.origin, offset );

	VectorNormalize( dir );
	
	if( *old_damage == 0 )
	{
		VectorCopy( offset, old_point );
		VectorCopy( dir, old_dir );
		*old_damage = damage;
		return;
	}

	frac = damage / (damage + *old_damage);
	for( i = 0; i < 3; i++ ) 
	{
		old_point[i] = (old_point[i] * (1.0f - frac)) + offset[i] * frac;
		old_dir[i] = (old_dir[i] * (1.0f - frac)) + dir[i] * frac;
	}
	*old_damage += damage;
}


void T_Damage (edict_t *targ, edict_t *inflictor, edict_t *attacker, vec3_t dir, vec3_t point, vec3_t normal, int damage, int knockback, int dflags, int mod)
{
	gclient_t	*client;
	int			take;
	int			save;
	int			asave;
	int			ev_sparks;

	if (!targ->takedamage)
		return;

	// friendly fire avoidance
	// if enabled you can't hurt teammates (but you can hurt yourself)
	// knockback still occurs
	if( G_IsTeamDamage(targ, attacker) )
		mod |= MOD_FRIENDLY_FIRE;
	else
		meansOfDeath = mod;

	client = targ->r.client;

	if (dflags & DAMAGE_BULLET)
		ev_sparks = EV_BULLET_SPARKS;
	else
		ev_sparks = EV_SPARKS;

	VectorNormalize(dir);

// bonus damage for suprising a monster
	if (!(dflags & DAMAGE_RADIUS) && (targ->r.svflags & SVF_MONSTER) && (attacker->r.client) && (!targ->enemy) && (targ->health > 0))
		damage *= 2;

	if (targ->flags & FL_NO_KNOCKBACK)
		knockback = 0;

// figure momentum add
	if (!(dflags & DAMAGE_NO_KNOCKBACK))
	{
		if ((knockback) && (targ->movetype != MOVETYPE_NONE) && (targ->movetype != MOVETYPE_BOUNCE) && (targ->movetype != MOVETYPE_PUSH) && (targ->movetype != MOVETYPE_STOP))
		{
			// Kurim knockback
       		float	scale;
			int		t;

			t = 2 * knockback;
			clamp( t, 50, 200 );

			if (targ->r.client)
				targ->r.client->ps.pmove.stats[PM_STAT_KNOCKBACK] = t;

			if (targ->mass < 50)
				scale = (float)knockback / 50;
			else
				scale = (float)knockback / targ->mass;

			if( targ->r.client && attacker == targ || game.gametype == GAMETYPE_MIDAIR )
				scale *= 1500.0;	// the rocket jump hack...
			else
				scale *= 1000.0;

			VectorMA (targ->velocity, scale, dir, targ->velocity);

/*			float	scale;

			if (targ->mass < 50)
				scale = (float)knockback / 50;
			else
				scale = (float)knockback / targ->mass;

			if (targ->r.client && attacker == targ)
				scale *= 1600.0;	// the rocket jump hack...
			else
				scale *= 500.0;

			VectorMA (targ->velocity, scale, dir, targ->velocity);
*/		}
	}

	// only apply knockback to...
	if ( game.gametype == GAMETYPE_RACE ||
		(game.gametype == GAMETYPE_MIDAIR && dflags & DAMAGE_RADIUS) )
		return;

	take = damage;
	save = 0;

	// check for godmode
	if ( (targ->flags & FL_GODMODE) && !(dflags & DAMAGE_NO_PROTECTION) )
	{
		take = 0;
		save = damage;

		//this spawned soft-blood particles
		//SpawnDamage (ev_sparks, point, normal, save);

	}
	
	asave = CheckArmor (targ, point, normal, take, ev_sparks, dflags);
	take -= asave;

	//treat cheat/powerup savings the same as armor
	asave += save;

	// team damage avoidance
	if( G_IsTeamDamage(targ, attacker) && !G_Gametype_CanTeamDamage(dflags) )
		return;

//ZOID
	//newgametypes
	G_Gametype_CTF_CheckHurtCarrier(targ, attacker);
//ZOID

// do the damage
	if (take)
	{
		//G_SIMPLE_DAMAGE_FEEDBACK [start]
		vec3_t dorigin, ddir;
		
		if( attacker )
			VectorSubtract( targ->s.origin, attacker->s.origin, ddir ); //just use last by now
		else if( inflictor )
			VectorSubtract( targ->s.origin, inflictor->s.origin, ddir ); //just use last by now
		else
			VectorCopy( normal, ddir );

		if( point[0] != 0.0f || point[1] != 0.0f || point[2] != 0.0f )
			VectorCopy( point, dorigin );
		else
			VectorSet( dorigin,
						targ->s.origin[0], 
						targ->s.origin[1],
						targ->s.origin[2] + targ->viewheight );

		G_BlendFrameDamage( targ, take, &targ->frame_damage_taken, dorigin, ddir, targ->frame_damage_at, targ->frame_damage_dir );
		G_BlendFrameDamage( targ, save, &targ->frame_damage_saved, dorigin, ddir, targ->frame_damage_at, targ->frame_damage_dir );
		//[end] of simple damage feedback

		targ->health = targ->health - take;

		// wsw : jal : accumulated damage for color blends
		if( targ->r.client ) {
			targ->r.client->damage_taken += take;
			targ->r.client->damage_saved += save;
		}

		if ( (take) && targ != attacker && targ->r.client && !targ->deadflag )
		{
			int hit;
			if (targ->health < 25)
				hit = PAIN_25;
			else if (targ->health < 50)
				hit = PAIN_50;
			else if (targ->health < 75)
				hit = PAIN_75;
			else
				hit = PAIN_100;

			if( attacker && attacker->r.client )
				attacker->r.client->damage_given += take;
		}
			
		if (targ->health <= 0)
		{
			if ((targ->r.svflags & SVF_MONSTER) || (client))
				targ->flags |= FL_NO_KNOCKBACK;
			Killed (targ, inflictor, attacker, take, point, mod);
			return;
		}
	}

	if (client)
	{
		if (!(targ->flags & FL_GODMODE) && (take) )
			targ->pain (targ, attacker, knockback, take);
	}
	else if (take)
	{
		if (targ->pain)
			targ->pain (targ, attacker, knockback, take);
	}
}


/*
============
T_RadiusDamage
============
*/
void T_RadiusDamage (edict_t *inflictor, edict_t *attacker, float maxdamage, float maxknockback, float mindamage, edict_t *ignore, float radius, int mod)
{
	float	points;
	edict_t	*ent = NULL;
	vec3_t	v;
	vec3_t	dir;
	int knockback;

	if( radius <= 0 )
		return;

	while ((ent = findradius(ent, inflictor->s.origin, radius)) != NULL)
	{
		if (ent == ignore)
			continue;
		if (!ent->takedamage)
			continue;

		VectorAdd (ent->r.mins, ent->r.maxs, v);
		VectorMA (ent->s.origin, 0.5, v, v);
		VectorSubtract (inflictor->s.origin, v, v);
		points = maxdamage - (maxdamage - mindamage) * (VectorLength (v) / radius);
		knockback = maxknockback * (1.0 - (VectorLength (v) / radius));
		if (ent == attacker)
			points = points * 0.5;
		if (points > 0)
		{
			if (CanDamage (ent, inflictor))
			{
				VectorSubtract (ent->s.origin, inflictor->s.origin, dir);
				T_Damage (ent, inflictor, attacker, dir, inflictor->s.origin, vec3_origin, (int)points, knockback, DAMAGE_RADIUS, mod);
			}
		}
	}
}
