/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "g_local.h"
#include "g_gametypes.h"


//==========================================================
//					Teams
//==========================================================

cvar_t	*g_maxteams;
cvar_t	*g_teams_maxplayers;
cvar_t	*g_teams_allow_lock;
cvar_t	*g_teams_teamdamage;

static char *g_teamNames[] =
{
	"SPECTATOR",
	"PLAYERS",
	"RED",
	"BLUE",
	"GREEN",
	"YELLOW",
	NULL
};

static char *g_teamSkinsNames[] =
{
	NULL,		//null means user defined skin
	NULL,
	"red",
	"blue",
	"green",
	"yellow",
	NULL
};

static qboolean Locked[GS_MAX_TEAMS];


//=================
//G_Teams_NewMap
//=================
void G_Teams_NewMap( void )
{
	int	team;

	//unlock all teams and clear up team lists
	memset( Locked, qfalse, sizeof(Locked) );
	memset( teamlist, 0, sizeof(teamlist) );
	for( team = TEAM_SPECTATOR; team < GS_MAX_TEAMS; team++ )
		teamlist[team].playerIndices[0] = -1;
}

//=================
//G_Teams_Init
//=================
void G_Teams_Init( void )
{
	//not much to do here, just start up the cvars
	g_maxteams = trap_Cvar_Get ("g_maxteams", "2", CVAR_ARCHIVE);
	if( g_maxteams->integer > GS_MAX_TEAMS - TEAM_PLAYERS ) {
		trap_Cvar_Set( "g_maxteams", va("%i",(GS_MAX_TEAMS - TEAM_PLAYERS)) );
	}

	g_teams_maxplayers = trap_Cvar_Get ("g_teams_maxplayers", "4", CVAR_ARCHIVE);
	g_teams_allow_lock = trap_Cvar_Get ("g_teams_allow_lock", "1", CVAR_ARCHIVE);
	g_teams_teamdamage = trap_Cvar_Get ("g_teams_teamdamage", "1", CVAR_ARCHIVE);
}


//=================
//G_Teams_JoinAnyTeam
// It's better to count the list in detail once per fame, than
// creating a quick list each time we need it.
//=================
void G_Teams_UpdateMembersList( void )
{
	static int			list[MAX_CLIENTS];
	static int			sorted[MAX_CLIENTS];
	static int			count;

	edict_t	*ent;
	int	i, team;
	int	bestscore;
	int	bestplayer = 0;

	for( team = TEAM_SPECTATOR; team < TEAM_RED + g_maxteams->integer; team++ )
	{
		count = 0;
		//create a temp list with the clients inside this team
		for( i = 0, ent = game.edicts + 1; i < game.maxclients; i++, ent++ )
		{
			if( !ent->r.inuse || !ent->r.client ) continue;
			
			if( ent->gameTeam == team ) {
				list[count] = ENTNUM(ent);
				count++;
			}
		}

		teamlist[team].numplayers = 0;

		if( count )
		{
			memset( sorted, 0, sizeof(int)*MAX_CLIENTS);
			bestplayer = -2;
			while( bestplayer != -1 )
			{
				bestscore = -9999;
				bestplayer = -1;
				//now sort them by their score
				for( i = 0; i < count; i++ ) {
					if( !sorted[i] )
					{
						ent = game.edicts + list[i];
						if( match.scores[PLAYERNUM(ent)].score >= bestscore ) {
							bestplayer = i;
							bestscore = match.scores[PLAYERNUM(ent)].score;
						}
					}
				}
				
				if( bestplayer > -1 ) {
					sorted[bestplayer] = qtrue;
					teamlist[team].playerIndices[teamlist[team].numplayers] = list[bestplayer];
					teamlist[team].numplayers++;
					
				}
			}
		}

		//terminate the list with -1
		teamlist[team].playerIndices[teamlist[team].numplayers] = -1; 
	}
}

//=================
//G_Teams_LockTeam
//=================
void G_Teams_LockTeam( int team )
{
	if( team && team < GS_MAX_TEAMS )
		Locked[team] = g_teams_allow_lock->integer;
}

//=================
//G_Teams_UnLockTeam
//=================
void G_Teams_UnLockTeam( int team )
{
	if( team && team < GS_MAX_TEAMS )
		Locked[team] = qfalse;
}

//=================
//G_TeamName
//=================
char *G_TeamName( int team )
{
	if( team < 0 && team >= GS_MAX_TEAMS )
		return NULL;

	return g_teamNames[team];
}

//=================
//G_TeamSkinName
//=================
char *G_TeamSkinName( int team )
{
	if( team < 0 && team >= GS_MAX_TEAMS )
		return NULL;

	return g_teamSkinsNames[team];
}

//=================
//G_TeamFromTeamName
//=================
int G_TeamFromTeamName( char *teamname )
{
	char *s;
	int		i;

	if( !teamname || !teamname[0] )
		return -1;	//invalid

	for( i = 0; i < GS_MAX_TEAMS; i++ ) 
	{
		s = g_teamNames[i];
		if( !Q_stricmp( s, teamname ) ) 
			return i;
	}
	return -1;	//invalid
}

//=================
//G_Teams_AssignTeamSkin
//=================
void G_Teams_AssignTeamSkin(edict_t *ent, char *s)
{
	char	*p;
	char	t[64];
	char	*uskin;

	uskin = G_TeamSkinName(ent->gameTeam);
	if( !uskin || !uskin[0] ) { //NULLs in the list mean user defined skin
		uskin = Info_ValueForKey (ent->r.client->pers.userinfo, "skin");
	} else {
		
		Q_snprintfz(t, sizeof(t), "%s", s);
		if ((p = strchr(t, '/')) != NULL)
			p[1] = 0;
		else
			strcpy(t, DEFAULT_PLAYERMODEL);
	}

	trap_ConfigString (CS_PLAYERSKINS + PLAYERNUM(ent), va("%s\\%i\\%s%s", 
			ent->r.client->pers.netname, ent->r.client->pers.hand, t, uskin) );
}

//=================
//G_GameTypes_CanJoinTeam - Gametypes check. Moveme, changeme, somethingme
//=================
qboolean G_GameTypes_CanJoinTeam( int team )
{
	if( team < 0 || team >= GS_MAX_TEAMS ) {
		G_Printf( "WARNING: 'G_GameTypes_CanJoinTeam' parsing a unrecognized team value\n" );
		return qfalse;
	}
	
	if( team == TEAM_SPECTATOR )
		return qtrue;

	if( game.gametype == GAMETYPE_SP ) {
		if( team == TEAM_PLAYERS )
			return qtrue;
	} else if( game.gametype == GAMETYPE_DM ) {
		if( team == TEAM_PLAYERS )
			return qtrue;
	} else if( game.gametype == GAMETYPE_DUEL ) {
		if( team >= TEAM_RED && team <= TEAM_YELLOW )
			return qtrue;
	} else if( game.gametype == GAMETYPE_TDM ) {
		if( team >= TEAM_RED && team <= TEAM_YELLOW )
			return qtrue;
	} else if( game.gametype == GAMETYPE_CTF ) {
		if( team >= TEAM_RED && team <= TEAM_YELLOW )
			return qtrue;
	} else if( game.gametype == GAMETYPE_RACE ) {
		if( team == TEAM_PLAYERS )
			return qtrue;
	} else if( game.gametype == GAMETYPE_MIDAIR ) {
		if( team == TEAM_PLAYERS )
			return qtrue;
	}
	
	return qfalse;
}

void G_Gametype_CTF_CleanUpPlayerStats( edict_t *ent );
//=================
//G_Teams_JoinTeam - The actual thing
//=================
qboolean G_Teams_JoinTeam( edict_t *ent, int team )
{
	char		*s;
	char		userinfo[MAX_INFO_STRING];

	if( !ent->r.client )	//fixme: handle non-clients too
		return qfalse;

	if( G_GameTypes_CanJoinTeam( team ) == qfalse ) {
		
		G_PrintMsg( ent, PRINT_HIGH, "Can't join %s in %s\n", G_TeamName( team ),gametypes[game.gametype].name );
		return qfalse;
	}

	if( team - TEAM_PLAYERS > g_maxteams->integer ) {
		G_PrintMsg( ent, PRINT_HIGH, "Can't join %s. Only %i teams allowed\n", G_TeamName( team ), g_maxteams->integer );
		return qfalse;
	}

	//see if team is locked
	if( G_Gametype_IsTeamBased() && Locked[team] && team > TEAM_SPECTATOR ) {
		G_PrintMsg( ent, PRINT_HIGH, "Team %s is LOCKED\n", G_TeamName( team ) );
		return qfalse;
	}

	//see if team is full
	if( team > TEAM_PLAYERS ) 
	{
		int	count = teamlist[team].numplayers;

		if( ( count + 1 > gametypes[game.gametype].maxPlayersPerTeam &&
			gametypes[game.gametype].maxPlayersPerTeam > 0 ) || 
			( count + 1 > g_teams_maxplayers->integer &&
			g_teams_maxplayers->integer > 0 ) ) {
			G_PrintMsg( ent, PRINT_HIGH, "Team %s is FULL\n", G_TeamName( team ) );
			return qfalse;
		}
	}

	//ok, can join, proceed
	
	//clean scores at changing team
	memset( &match.scores[PLAYERNUM(ent)], 0, sizeof(client_scores_t) );
	if( game.gametype == GAMETYPE_CTF ) {
		G_Gametype_CTF_DeadDropFlag(ent);
		G_Gametype_CTF_CleanUpPlayerStats(ent);
	}

	if( team == TEAM_SPECTATOR ) {

		ent->deadflag = DEAD_NO;
		ent->movetype = MOVETYPE_NOCLIP;
		ent->r.solid = SOLID_NOT;
		ent->r.svflags |= SVF_NOCLIENT;

		ent->gameTeam = TEAM_SPECTATOR;	//force as team spectator   
		memcpy (userinfo, ent->r.client->pers.userinfo, sizeof(userinfo));
		InitClientPersistant(ent->r.client);
		ClientUserinfoChanged (ent, userinfo);
		G_ClearPlayerStateEvents( ent->r.client );
	
		G_AddEvent (ent, EV_TELEPORT, 0, qtrue);
		trap_LinkEntity (ent);

	} else { //known team found

		ent->gameTeam = team;
		s = Info_ValueForKey (ent->r.client->pers.userinfo, "skin");
		G_Teams_AssignTeamSkin(ent, s);

		// call for spawn
		ent->movetype = MOVETYPE_NOCLIP; // so respawn doesn't create a corpse
		G_Gametype_ClientRespawn(ent);
	}

	G_Teams_UpdateMembersList();
	return qtrue;
}

//=================
//G_Teams_JoinAnyTeam - find us a team since we are too lazy to do ourselves
//=================
void G_Teams_JoinAnyTeam( edict_t *ent )
{
	int		best = -1;
	int		i, team;

	//depending on the gametype, of course

	if( !G_Gametype_IsTeamBased() ) {
		if( ent->gameTeam == TEAM_PLAYERS ) {
			G_PrintMsg( ent, PRINT_HIGH, "You are already in %s team\n", G_TeamName(TEAM_PLAYERS) );
			return;
		}
		if( G_Teams_JoinTeam( ent, TEAM_PLAYERS ) )
			G_PrintMsg (NULL, PRINT_HIGH, "%s%s joined the %s team.\n",
			ent->r.client->pers.netname, S_COLOR_WHITE, G_TeamName(ent->gameTeam) );
		return;

	} else {	//team based

		//find the available team with smaller player count
		for( i = TEAM_PLAYERS; i < TEAM_RED + g_maxteams->integer; i++ )
		{
			if( G_GameTypes_CanJoinTeam(i) == qfalse )
				continue;

			if( best == -1 || teamlist[i].numplayers < best ) {
				best = teamlist[i].numplayers;
				team = i;
			}
		}

		if( team == ent->gameTeam ) { // he is at the right team
			G_PrintMsg (ent, PRINT_HIGH, "%sCouldn't find an emptier team than team %s.\n",
				S_COLOR_WHITE, G_TeamName(ent->gameTeam) );
			return;
		}

		if( best != -1 ) {
			if( G_Teams_JoinTeam( ent, team ) )
				G_PrintMsg (NULL, PRINT_HIGH, "%s%s joined the %s team.\n",
				ent->r.client->pers.netname, S_COLOR_WHITE, G_TeamName(ent->gameTeam) );
			return;
		}
	}

	G_PrintMsg( ent, PRINT_HIGH, "You can't join the game now\n" );
}

//=================
//G_Teams_Join_Cmd
//=================
void G_Teams_Join_Cmd( edict_t *ent )
{
	char	*wishteam, *t;
	int		team;

	t = trap_Cmd_Argv(1);	//should be 'team', 'game' or nothing
	if( !t || *t == 0 || !Q_stricmp( "game", t ) ) {
		G_Teams_JoinAnyTeam(ent);
		return;
	}

	//try the shortcut (join <teamname>)
	team = G_TeamFromTeamName( t );
	if( team != -1 ) {
		if( team == ent->gameTeam ) {
			G_PrintMsg( ent, PRINT_HIGH, "You are already in %s team\n", G_TeamName(team) );
			return;
		}
		if( G_Teams_JoinTeam( ent, team ) ) //found a team to join
			G_PrintMsg (NULL, PRINT_HIGH, "%s%s joined the %s team.\n",
			ent->r.client->pers.netname, S_COLOR_WHITE, G_TeamName(ent->gameTeam) );
		return;
	}

	//try the extended format (join team <teamname>)
	if( !Q_stricmp( "team", t ) ) {

		wishteam = trap_Cmd_Argv(2);
		team = G_TeamFromTeamName( wishteam );
		if( team != -1 ) {
			if( team == ent->gameTeam ) {
				G_PrintMsg( ent, PRINT_HIGH, "You are already in %s team\n", G_TeamName(team) );
				return;
			}
			if( G_Teams_JoinTeam( ent, team ) ) //found a team to join
				G_PrintMsg (NULL, PRINT_HIGH, "%s%s joined the %s team.\n",
				ent->r.client->pers.netname, S_COLOR_WHITE, G_TeamName(ent->gameTeam) );
			return;
		}
	}

	G_PrintMsg( ent, PRINT_HIGH, "Wrong join option. Use 'join', 'join <team name>', 'join team <teamname> or 'join game' for any team\n" );
}


//======================================================================
//
//TEAM COMMUNICATIONS
//
//======================================================================

//jal: I'm not sure if we should keep this player seen feature

static void loc_buildboxpoints(vec3_t p[8], vec3_t org, vec3_t mins, vec3_t maxs)
{
	VectorAdd(org, mins, p[0]);
	VectorCopy(p[0], p[1]);
	p[1][0] -= mins[0];
	VectorCopy(p[0], p[2]);
	p[2][1] -= mins[1];
	VectorCopy(p[0], p[3]);
	p[3][0] -= mins[0];
	p[3][1] -= mins[1];
	VectorAdd(org, maxs, p[4]);
	VectorCopy(p[4], p[5]);
	p[5][0] -= maxs[0];
	VectorCopy(p[0], p[6]);
	p[6][1] -= maxs[1];
	VectorCopy(p[0], p[7]);
	p[7][0] -= maxs[0];
	p[7][1] -= maxs[1];
}

qboolean loc_CanSee (edict_t *targ, edict_t *inflictor)
{
	trace_t	trace;
	vec3_t	targpoints[8];
	int i;
	vec3_t viewpoint;

// bmodels need special checking because their origin is 0,0,0
	if (targ->movetype == MOVETYPE_PUSH)
		return qfalse; // bmodels not supported

	loc_buildboxpoints(targpoints, targ->s.origin, targ->r.mins, targ->r.maxs);
	
	VectorCopy(inflictor->s.origin, viewpoint);
	viewpoint[2] += inflictor->viewheight;

	for (i = 0; i < 8; i++) {
		trap_Trace (&trace, viewpoint, vec3_origin, vec3_origin, targpoints[i], inflictor, MASK_SOLID);
		if (trace.fraction == 1.0)
			return qtrue;
	}

	return qfalse;
}

//==========================================================
//		SAY TEAM
//==========================================================
//
static void CTFSay_Team_Location(edict_t *who, char *buf, int buflen)
{
	edict_t *what = NULL;
	edict_t *hot = NULL;
	float hotdist = 3.0f*8192.0f*8192.0f, newdist;
	vec3_t v;

	while ((what = G_Find(what, FOFS(classname), "target_location")) != NULL) {
		VectorSubtract( what->s.origin, who->s.origin, v );
		newdist = VectorLength( v );

		if ( newdist > hotdist ) {
			continue;
		}

		if ( !trap_inPVS ( what->s.origin, who->s.origin ) )
			continue;

		hot = what;
		hotdist = newdist;
	}

	if (!hot) {
		buf[0] = 0;
		return;
	}

	// we now have the closest target_location
	if ( hot->count ) {
		Q_snprintfz ( buf, buflen, "%c%c%s" S_COLOR_WHITE, Q_COLOR_ESCAPE, hot->count + '0', hot->message );
	} else {
		Q_snprintfz ( buf, buflen, "%s", hot->message );
	}
}

static void CTFSay_Team_Armor(edict_t *who, char *buf, int buflen)
{
	gitem_t		*item;
	int			armor_tag = 0;

	*buf = 0;

	if( who->r.client )
		armor_tag = ArmorTag (who->r.client->pers.inventory);
	if (armor_tag)
	{
		item = game.items[armor_tag];
		if (item) {
			if (*buf)
				strcat(buf, "and ");
			sprintf(buf+strlen(buf), "%i units of %s",
				who->r.client->pers.inventory[armor_tag], item->pickup_name);
		}
	}

	if (!*buf)
		Q_snprintfz(buf, buflen, "no armor");
}

static void CTFSay_Team_Health(edict_t *who, char *buf, int buflen)
{
	if (who->health <= 0)
		Q_snprintfz (buf, buflen, "dead");
	else
		Q_snprintfz (buf, buflen, "%i health", who->health);
}



static void CTFSay_Team_Weapon(edict_t *who, char *buf, int buflen)
{
	if (who->s.weapon)
		Q_snprintfz(buf, buflen, game.items[who->s.weapon]->pickup_name);
	else
		Q_snprintfz(buf, buflen, "none");
}

static void CTFSay_Team_Sight(edict_t *who, char *buf, int buflen)
{
	int i;
	edict_t *targ;
	int n = 0;
	char s[1024];
	char s2[1024];

	*s = *s2 = 0;
	for (i = 1; i <= game.maxclients; i++) {
		targ = game.edicts + i;
		if (!targ->r.inuse || 
			targ == who ||
			!loc_CanSee(targ, who))
			continue;
		if (*s2) {
			if (strlen(s) + strlen(s2) + 3 < sizeof(s)) {
				if (n)
					strcat(s, ", ");
				strcat(s, s2);
				*s2 = 0;
			}
			n++;
		}
		strcpy(s2, targ->r.client->pers.netname);
	}
	if (*s2) {
		if (strlen(s) + strlen(s2) + 6 < sizeof(s)) {
			if (n)
				strcat(s, " and ");
			strcat(s, s2);
		}
		Q_snprintfz(buf, buflen, s);
	} else
		Q_snprintfz(buf, buflen, "no one");
}

void CTFSay_Team(edict_t *who, char *msg)
{
	char outmsg[256];
	char buf[256];
	int i;
	char *p;
	edict_t *cl_ent;

	if (CheckFlood(who))
		return;

	outmsg[0] = 0;

	if (*msg == '\"') {
		msg[strlen(msg) - 1] = 0;
		msg++;
	}

	for (p = outmsg; *msg && (p - outmsg) < sizeof(outmsg) - 2; msg++) {
		if (*msg == '%') {
			switch (*++msg) {
				case 'l' :
				case 'L' :
					CTFSay_Team_Location(who, buf, sizeof(buf));
					if (strlen(buf) + (p - outmsg) < sizeof(outmsg) - 2) {
						if ( buf[0] ) {
							strcpy(p, buf);
							p += strlen(buf);
						}
					}
					break;
				case 'a' :
				case 'A' :
					CTFSay_Team_Armor(who, buf, sizeof(buf));
					if (strlen(buf) + (p - outmsg) < sizeof(outmsg) - 2) {
						strcpy(p, buf);
						p += strlen(buf);
					}
					break;
				case 'h' :
				case 'H' :
					CTFSay_Team_Health(who, buf, sizeof(buf));
					if (strlen(buf) + (p - outmsg) < sizeof(outmsg) - 2) {
						strcpy(p, buf);
						p += strlen(buf);
					}
					break;

				case 'w' :
				case 'W' :
					CTFSay_Team_Weapon(who, buf, sizeof(buf));
					if (strlen(buf) + (p - outmsg) < sizeof(outmsg) - 2) {
						strcpy(p, buf);
						p += strlen(buf);
					}
					break;

				case 'n' :
				case 'N' :
					CTFSay_Team_Sight(who, buf, sizeof(buf));
					if (strlen(buf) + (p - outmsg) < sizeof(outmsg) - 2) {
						strcpy(p, buf);
						p += strlen(buf);
					}
					break;

				default :
					*p++ = *msg;
			}
		} else
			*p++ = *msg;
	}
	*p = 0;

	for (i = 0; i < game.maxclients; i++) {
		cl_ent = game.edicts + 1 + i;
		if (!cl_ent->r.inuse)
			continue;
		if (cl_ent->gameTeam == who->gameTeam)
			G_PrintMsg (cl_ent, PRINT_CHAT, "%s%s: %s\n", 
				who->r.client->pers.netname, S_COLOR_MAGENTA, outmsg);
	}
}



