/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "g_local.h"
#include "g_gametypes.h"

/*QUAKED target_temp_entity (1 0 0) (-8 -8 -8) (8 8 8)
Fire an origin based temp entity event to the clients.
"style"		type byte
*/
void Use_Target_Tent (edict_t *ent, edict_t *other, edict_t *activator)
{
	G_SpawnEvent ( ent->style, 0, ent->s.origin );
}

void SP_target_temp_entity (edict_t *ent)
{
	ent->use = Use_Target_Tent;
}


//==========================================================

//==========================================================


/*QUAKED target_speaker (0 .7 .7) (-8 -8 -8) (8 8 8) LOOPED_ON LOOPED_OFF RELIABLE GLOBAL ACTIVATOR
Normal sounds play each time the target is used.  The reliable flag can be set for crucial voiceovers. Looped sounds are always atten 3 / vol 1, and the use function toggles it on/off. Multiple identical looping sounds will just increase volume without any speed cost.
-------- KEYS --------
noise : path/name of .wav file to play (eg. sound/world/growl1.wav - see Notes).
volume : 0.0 to 1.0
attenuation : -1 = none, send to whole level, 1 = normal fighting sounds, 2 = idle sound level, 3 = ambient sound level
wait : delay in seconds between each time the sound is played ("random" key must be set - see Notes).
random : random time variance in seconds added or subtracted from "wait" delay ("wait" key must be set - see Notes).
targetname : the activating button or trigger points to this.
notsingle : when set to 1, entity will not spawn in Single Player mode
notfree : when set to 1, entity will not spawn in "Free for all" and "Tournament" modes.
notduel : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
notteam : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. 
notctf : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
-------- SPAWNFLAGS --------
LOOPED_ON : &1 sound will loop and initially start on in level (will toggle on/off when triggered).
LOOPED_OFF : &2 sound will loop and initially start off in level (will toggle on/off when triggered).
RELIABLE : &4
GLOBAL : &8 Overrides attenuation setting. Sound will play full volume throughout the level as if it had attenuation -1
ACTIVATOR : &16 sound will play only for the player that activated the target.
-------- NOTES --------
The path portion value of the "noise" key can be replaced by the implicit folder character "*" for triggered sounds that belong to a particular player model. For example, if you want to create a "bottomless pit" in which the player screams and dies when he falls into, you would place a trigger_multiple over the floor of the pit and target a target_speaker with it. Then, you would set the "noise" key to "*falling1.wav". The * character means the current player model's sound folder. So if your current player model is Visor, * = sound/player/visor, if your current player model is Sarge, * = sound/player/sarge, etc. This cool feature provides an excellent way to create "player-specific" triggered sounds in your levels.*/


void Use_Target_Speaker (edict_t *ent, edict_t *other, edict_t *activator)
{
	if (ent->spawnflags & 3)
	{	// looping sound toggles
		if (ent->s.sound)
			ent->s.sound = 0;	// turn it off
		else
			ent->s.sound = ent->noise_index;	// start it
	}
	else
	{	// normal sound
		if( ent->spawnflags & 8 )
			G_Sound( activator, CHAN_VOICE|CHAN_RELIABLE, ent->noise_index, ent->volume, ent->attenuation );
		// use a G_PositionedSound, because this entity won't normally be
		// sent to any clients because it is invisible
		else if( ent->spawnflags & 4 )
			G_PositionedSound( ent->s.origin, ent, CHAN_VOICE|CHAN_RELIABLE, ent->noise_index, ent->volume, ent->attenuation );
		else
			G_PositionedSound( ent->s.origin, ent, CHAN_VOICE, ent->noise_index, ent->volume, ent->attenuation );
	}
}

void SP_target_speaker (edict_t *ent)
{
	char	buffer[MAX_QPATH];

	if (!st.noise)
	{
		if (developer->integer)
			G_Printf ("target_speaker with no noise set at %s\n", vtos(ent->s.origin));
		return;
	}

	if (!strstr (st.noise, ".wav"))
		Q_snprintfz (buffer, sizeof(buffer), "%s.wav", st.noise);
	else
		Q_strncpyz (buffer, st.noise, sizeof(buffer));
	ent->noise_index = trap_SoundIndex (buffer);

	if (!ent->volume)
		ent->volume = 1.0;
	
	if (ent->attenuation == -1 || ent->spawnflags & 8)	// use -1 so 0 defaults to ATTN_NONE
		ent->attenuation = ATTN_NONE;
	else if (!ent->attenuation)
		ent->attenuation = ATTN_NORM;

	// check for prestarted looping sound
	if (ent->spawnflags & 1)
		ent->s.sound = ent->noise_index;

	ent->use = Use_Target_Speaker;

	// must link the entity so we get areas and clusters so
	// the server can determine who to send updates to
	trap_LinkEntity (ent);
}


//==========================================================

/*QUAKED target_help (1 0 1) (-16 -16 -24) (16 16 24) help1
When fired, the "message" key becomes the current personal computer string, and the message light will be set on all clients status bars.
*/
void SP_target_help(edict_t *ent)
{
	// jal: removedcmdhelp
	G_FreeEdict (ent);
}

//==========================================================

/*QUAKED target_secret (1 0 1) (-8 -8 -8) (8 8 8)
Counts a secret found.
These are single use targets.
*/
void use_target_secret (edict_t *ent, edict_t *other, edict_t *activator)
{
	G_Sound (ent, CHAN_VOICE, ent->noise_index, 1, ATTN_NORM);

	level.found_secrets++;

	G_UseTargets (ent, activator);
	G_FreeEdict (ent);
}

void SP_target_secret (edict_t *ent)
{
	if (game.gametype != GAMETYPE_SP /*&& game.gametype != GAMETYPE_COOP*/ )	//newgametypes
	{	// auto-remove for deathmatch
		G_FreeEdict (ent);
		return;
	}

	ent->use = use_target_secret;
	if (!st.noise)
		st.noise = "sound/misc/secret.wav";
	ent->noise_index = trap_SoundIndex (st.noise);
	ent->r.svflags = SVF_NOCLIENT;
	level.total_secrets++;
}

//==========================================================

/*QUAKED target_goal (1 0 1) (-8 -8 -8) (8 8 8)
Counts a goal completed.
These are single use targets.
*/
void use_target_goal (edict_t *ent, edict_t *other, edict_t *activator)
{
	G_Sound (ent, CHAN_VOICE, ent->noise_index, 1, ATTN_NORM);

	level.found_goals++;

	G_UseTargets (ent, activator);
	G_FreeEdict (ent);
}

void SP_target_goal (edict_t *ent)
{
	if (game.gametype != GAMETYPE_SP /*&& game.gametype != GAMETYPE_COOP*/)//newgametypes
	{	// auto-remove for deathmatch
		G_FreeEdict (ent);
		return;
	}

	ent->use = use_target_goal;
	if (!st.noise)
		st.noise = "sound/misc/secret.wav";
	ent->noise_index = trap_SoundIndex (st.noise);
	ent->r.svflags = SVF_NOCLIENT;
	level.total_goals++;
}

//==========================================================


/*QUAKED target_explosion (1 0 0) (-8 -8 -8) (8 8 8)
Spawns an explosion temporary entity when used.

"delay"		wait this long before going off
"dmg"		how much radius damage should be done, defaults to 0
*/
void target_explosion_explode (edict_t *self)
{
	float		save;
	
	G_SpawnEvent ( EV_EXPLOSION1, 0, self->s.origin );
	T_RadiusDamage (self, self->activator, self->dmg, self->dmg, 0, NULL, self->dmg+40, MOD_EXPLOSIVE);

	save = self->delay;
	self->delay = 0;
	G_UseTargets (self, self->activator);
	self->delay = save;
}

void use_target_explosion (edict_t *self, edict_t *other, edict_t *activator)
{
	self->activator = activator;

	if (!self->delay)
	{
		target_explosion_explode (self);
		return;
	}

	self->think = target_explosion_explode;
#ifdef SV_FRAMETIME
	self->nextthink = level.timemsec + self->delay * 1000;
#else
	self->nextthink = level.time + self->delay;
#endif
}

void SP_target_explosion (edict_t *ent)
{
	ent->use = use_target_explosion;
	ent->r.svflags = SVF_NOCLIENT;
}


//==========================================================

/*QUAKED target_changelevel (1 0 0) (-8 -8 -8) (8 8 8)
Changes level to "map" when fired
*/
void use_target_changelevel (edict_t *self, edict_t *other, edict_t *activator)
{
	if (match.state >= MATCH_STATE_POSTMATCH)
		return;		// already activated

	if (game.gametype != GAMETYPE_SP /*&& game.gametype != GAMETYPE_COOP*/)//newgametypes
	{
		if (game.edicts[1].health <= 0)
			return;
	}

	// if noexit, do a ton of damage to other
	if (game.gametype != GAMETYPE_SP /*&& game.gametype != GAMETYPE_COOP*/ && !( dmflags->integer & DF_ALLOW_EXIT) && other != world) //newgametypes
	{
		T_Damage (other, self, self, vec3_origin, other->s.origin, vec3_origin, 10 * other->max_health, 1000, 0, MOD_EXIT);
		return;
	}

	// if multiplayer, let everyone know who hit the exit
	if (game.gametype != GAMETYPE_SP /*&& game.gametype != GAMETYPE_COOP*/)//newgametypes
	{
		if (activator && activator->r.client)
			G_PrintMsg (NULL, PRINT_HIGH, "%s exited the level.\n", activator->r.client->pers.netname);
	}

	// if going to a new unit, clear cross triggers
	if (strstr(self->map, "*"))	
		game.serverflags &= ~(SFL_CROSS_TRIGGER_MASK);

//	BeginIntermission (self);
	G_EndMatch();
}

void SP_target_changelevel (edict_t *ent)
{
	if (!ent->map)
	{
		if (developer->integer)
			G_Printf ("target_changelevel with no map at %s\n", vtos(ent->s.origin));
		G_FreeEdict (ent);
		return;
	}

	ent->use = use_target_changelevel;
	ent->r.svflags = SVF_NOCLIENT;
}


//==========================================================

/*QUAKED target_splash (1 0 0) (-8 -8 -8) (8 8 8)
Creates a particle splash effect when used.

"count"	how many pixels in the splash
"dmg"	if set, does a radius damage at this location when it splashes
		useful for lava/sparks
"color" color in r g b floating point format
*/

void use_target_splash (edict_t *self, edict_t *other, edict_t *activator)
{
	edict_t *event;

	event = G_SpawnEvent ( EV_LASER_SPARKS, DirToByte (self->movedir), self->s.origin );
	event->s.eventCount = self->count;

	// default to none
	if (VectorCompare (self->color, vec3_origin))
		event->s.colorRGBA = 0;
	else 
		event->s.colorRGBA = COLOR_RGBA (
		(int)(self->color[0]*255)&255,
		(int)(self->color[1]*255)&255,
		(int)(self->color[2]*255)&255, 
		255);

	if (self->dmg)
		T_RadiusDamage (self, activator, self->dmg, self->dmg, 0, NULL, self->dmg+40, MOD_SPLASH);
}

void SP_target_splash (edict_t *self)
{
	self->use = use_target_splash;
	G_SetMovedir (self->s.angles, self->movedir);

	if (!self->count)
		self->count = 32;

	self->r.svflags = SVF_NOCLIENT;
}


//==========================================================

/*QUAKED target_spawner (1 0 0) (-8 -8 -8) (8 8 8)
Set target to the type of entity you want spawned.
Useful for spawning monsters and gibs in the factory levels.

For monsters:
	Set direction to the facing you want it to have.

For gibs:
	Set direction if you want it moving and
	speed how fast it should be moving otherwise it
	will just be dropped
*/
void ED_CallSpawn (edict_t *ent);

void use_target_spawner (edict_t *self, edict_t *other, edict_t *activator)
{
	edict_t	*ent;

	ent = G_Spawn();
	ent->classname = self->target;
	VectorCopy (self->s.origin, ent->s.origin);
	VectorCopy (self->s.angles, ent->s.angles);
	ED_CallSpawn (ent);
	trap_UnlinkEntity (ent);
	KillBox (ent);
	trap_LinkEntity (ent);
	if (self->speed)
		VectorCopy (self->movedir, ent->velocity);
}

void SP_target_spawner (edict_t *self)
{
	self->use = use_target_spawner;
	self->r.svflags = SVF_NOCLIENT;
	if (self->speed)
	{
		G_SetMovedir (self->s.angles, self->movedir);
		VectorScale (self->movedir, self->speed, self->movedir);
	}
}

//==========================================================

/*QUAKED target_blaster (1 0 0) (-8 -8 -8) (8 8 8) NOTRAIL NOEFFECTS
Fires a blaster bolt in the set direction when triggered.

dmg		default is 15
speed	default is 1000
*/

void W_Fire_Plasma (edict_t *self, vec3_t start, vec3_t dir, int damage, int radius_damage, int radius, int speed, int mod );
void use_target_blaster (edict_t *self, edict_t *other, edict_t *activator)
{
	int type;

	if (self->spawnflags & 2)
		type = ET_GENERIC;
	else if (self->spawnflags & 1)
		type = ET_ELECTRO_WEAK;
	else
		type = ET_BLASTER;

	//fire_blaster (self, self->s.origin, self->movedir, self->dmg, self->speed, type, MOD_TARGET_BLASTER);
	W_Fire_Plasma (self, self->s.origin, self->movedir, self->dmg, self->dmg, 70, self->speed, MOD_TARGET_BLASTER );
	G_Sound (self, CHAN_VOICE, self->noise_index, 1, ATTN_NORM);
}

void SP_target_blaster (edict_t *self)
{
	self->use = use_target_blaster;
	G_SetMovedir (self->s.angles, self->movedir);
	self->noise_index = trap_SoundIndex ("sound/weapons/laser2.wav");

	if (!self->dmg)
		self->dmg = 15;
	if (!self->speed)
		self->speed = 1000;

	self->r.svflags = SVF_NOCLIENT;
}


//==========================================================

/*QUAKED target_crosslevel_trigger (.5 .5 .5) (-8 -8 -8) (8 8 8) trigger1 trigger2 trigger3 trigger4 trigger5 trigger6 trigger7 trigger8
Once this trigger is touched/used, any trigger_crosslevel_target with the same trigger number is automatically used when a level is started within the same unit.  It is OK to check multiple triggers.  Message, delay, target, and killtarget also work.
*/
void trigger_crosslevel_trigger_use (edict_t *self, edict_t *other, edict_t *activator)
{
	game.serverflags |= self->spawnflags;
	G_FreeEdict (self);
}

void SP_target_crosslevel_trigger (edict_t *self)
{
	self->r.svflags = SVF_NOCLIENT;
	self->use = trigger_crosslevel_trigger_use;
}

/*QUAKED target_crosslevel_target (.5 .5 .5) (-8 -8 -8) (8 8 8) trigger1 trigger2 trigger3 trigger4 trigger5 trigger6 trigger7 trigger8
Triggered by a trigger_crosslevel elsewhere within a unit.  If multiple triggers are checked, all must be true.  Delay, target and
killtarget also work.

"delay"		delay before using targets if the trigger has been activated (default 1)
*/
void target_crosslevel_target_think (edict_t *self)
{
	if (self->spawnflags == (game.serverflags & SFL_CROSS_TRIGGER_MASK & self->spawnflags))
	{
		G_UseTargets (self, self);
		G_FreeEdict (self);
	}
}

void SP_target_crosslevel_target (edict_t *self)
{
	if (! self->delay)
		self->delay = 1;
	self->r.svflags = SVF_NOCLIENT;

	self->think = target_crosslevel_target_think;
#ifdef SV_FRAMETIME
	self->nextthink = level.timemsec + self->delay * 1000;
#else
	self->nextthink = level.time + self->delay;
#endif
}

//==========================================================

/*
QUAKED target_laser (0 .5 0) (-8 -8 -8) (8 8 8) START_ON RED GREEN BLUE YELLOW ORANGE FAT
When triggered, fires a laser.  You can either set a target or a direction.
-------- KEYS --------
angles: alternate "pitch, yaw, roll" angles method of aiming laser (default 0 0 0).
target : point this to a target_position entity to set the laser's aiming direction.
targetname : the activating trigger points to this.
notsingle : when set to 1, entity will not spawn in Single Player mode
notfree : when set to 1, entity will not spawn in "Free for all" and "Tournament" modes.
notduel : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
notteam : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. 
notctf : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
-------- SPAWNFLAGS --------
START_ON : when set, the laser will start on in the game.
RED : 
GREEN : BLUE :
YELLOW :
ORANGE :
FAT :
*/
void target_laser_think (edict_t *self)
{
	edict_t	*ignore;
	vec3_t	start;
	vec3_t	end;
	trace_t	tr;
	vec3_t	point;
	vec3_t	last_movedir;
	int		count;

	// our lifetime has expired
	if (self->delay && (self->wait < level.time))
	{
		if ( self->r.owner && self->r.owner->use ) {
			self->r.owner->use ( self->r.owner, self, self->activator );
		}

		G_FreeEdict (self);
		return;
	}

	if (self->spawnflags & 0x80000000)
		count = 8;
	else
		count = 4;

	if (self->enemy)
	{
		VectorCopy (self->movedir, last_movedir);
		VectorMA (self->enemy->r.absmin, 0.5, self->enemy->r.size, point);
		VectorSubtract (point, self->s.origin, self->movedir);
		VectorNormalize (self->movedir);
		if (!VectorCompare(self->movedir, last_movedir))
			self->spawnflags |= 0x80000000;
	}

	ignore = self;
	VectorCopy (self->s.origin, start);
	VectorMA (start, 2048, self->movedir, end);
	while (1)
	{
		trap_Trace (&tr, start, NULL, NULL, end, ignore, MASK_SHOT);
		if (tr.fraction == 1)
			break;

		// hurt it if we can
		if ((game.edicts[tr.ent].takedamage) && !(game.edicts[tr.ent].flags & FL_IMMUNE_LASER))
		{
			if (game.edicts[tr.ent].r.client && self->activator->r.client)
			{
//newgametypes[start]
				if ( !G_Gametype_IsTeamBased() || 
					game.edicts[tr.ent].gameTeam != self->activator->gameTeam)
//newgametypes[end]
					T_Damage (&game.edicts[tr.ent], self, self->activator, self->movedir, tr.endpos, vec3_origin, self->dmg, 1, DAMAGE_ENERGY, self->count);
			}
			else
			{
				T_Damage (&game.edicts[tr.ent], self, self->activator, self->movedir, tr.endpos, vec3_origin, self->dmg, 1, DAMAGE_ENERGY, self->count);
			}
		}

		// if we hit something that's not a monster or player or is immune to lasers, we're done
		if (!(game.edicts[tr.ent].r.svflags & SVF_MONSTER) && (!game.edicts[tr.ent].r.client))
		{
			if (self->spawnflags & 0x80000000)
			{
				edict_t *event;

				self->spawnflags &= ~0x80000000;

				event = G_SpawnEvent ( EV_LASER_SPARKS, DirToByte (tr.plane.normal), tr.endpos );
				event->s.eventCount = count;
				event->s.colorRGBA = self->s.colorRGBA;
			}
			break;
		}

		ignore = &game.edicts[tr.ent];
		VectorCopy (tr.endpos, start);
	}

	VectorCopy (tr.endpos, self->s.origin2);
#ifdef SV_FRAMETIME
	self->nextthink = level.timemsec + G_FRAMETIME;
#else
	self->nextthink = level.time + FRAMETIME;
#endif
}

void target_laser_on (edict_t *self)
{
	if (!self->activator)
		self->activator = self;
	self->spawnflags |= 0x80000001;
	self->r.svflags &= ~SVF_NOCLIENT;
	self->wait = level.time + self->delay;
	target_laser_think (self);
}

void target_laser_off (edict_t *self)
{
	self->spawnflags &= ~1;
	self->r.svflags |= SVF_NOCLIENT;
	self->nextthink = 0;
}

void target_laser_use (edict_t *self, edict_t *other, edict_t *activator)
{
	self->activator = activator;
	if (self->spawnflags & 1)
		target_laser_off (self);
	else
		target_laser_on (self);
}

void target_laser_start (edict_t *self)
{
	edict_t *ent;

	self->movetype = MOVETYPE_NONE;
	self->r.solid = SOLID_NOT;
	self->s.type = ET_BEAM;
	self->s.modelindex = 1;			// must be non-zero
	self->r.svflags = SVF_FORCEOLDORIGIN;

	// set the beam diameter
	if (self->spawnflags & 64)
		self->s.frame = 16;
	else
		self->s.frame = 4;

	// set the color
	if (self->spawnflags & 2)
		self->s.colorRGBA = COLOR_RGBA (220, 0, 0, 76);
	else if (self->spawnflags & 4)
		self->s.colorRGBA = COLOR_RGBA (0, 220, 0, 76);
	else if (self->spawnflags & 8)
		self->s.colorRGBA = COLOR_RGBA (0, 0, 220, 76);
	else if (self->spawnflags & 16)
		self->s.colorRGBA = COLOR_RGBA (220, 220, 0, 76);
	else if (self->spawnflags & 32)
		self->s.colorRGBA = COLOR_RGBA (255, 255, 0, 76);

	if (!self->enemy)
	{
		if (self->target)
		{
			ent = G_Find (NULL, FOFS(targetname), self->target);
			if (!ent)
				if (developer->integer)
					G_Printf ("%s at %s: %s is a bad target\n", self->classname, vtos(self->s.origin), self->target);
			self->enemy = ent;
		}
		else
		{
			G_SetMovedir (self->s.angles, self->movedir);
		}
	}
	self->use = target_laser_use;
	self->think = target_laser_think;

	if (!self->dmg)
		self->dmg = 1;

	VectorSet (self->r.mins, -8, -8, -8);
	VectorSet (self->r.maxs, 8, 8, 8);
	trap_LinkEntity (self);

	if (self->spawnflags & 1)
		target_laser_on (self);
	else
		target_laser_off (self);
}

void SP_target_laser (edict_t *self)
{
	// let everything else get spawned before we start firing
	self->think = target_laser_start;
#ifdef SV_FRAMETIME
	self->nextthink = level.timemsec + 1000;
#else
	self->nextthink = level.time + 1;
#endif
	self->count = MOD_TARGET_LASER;
}

//==========================================================

/*QUAKED target_lightramp (0 .5 .8) (-8 -8 -8) (8 8 8) TOGGLE
speed		How many seconds the ramping will take
message		two letters; starting lightlevel and ending lightlevel
*/

void target_lightramp_think (edict_t *self)
{
	char	style[2];

	style[0] = 'a' + self->movedir[0] + (level.time - self->timestamp) / FRAMETIME * self->movedir[2];
	style[1] = 0;
	trap_ConfigString (CS_LIGHTS+self->enemy->style, style);

	if ((level.time - self->timestamp) < self->speed)
	{
#ifdef SV_FRAMETIME
		self->nextthink = level.timemsec + G_FRAMETIME;
#else
		self->nextthink = level.time + FRAMETIME;
#endif
	}
	else if (self->spawnflags & 1)
	{
		char	temp;

		temp = self->movedir[0];
		self->movedir[0] = self->movedir[1];
		self->movedir[1] = temp;
		self->movedir[2] *= -1;
	}
}

void target_lightramp_use (edict_t *self, edict_t *other, edict_t *activator)
{
	if (!self->enemy)
	{
		edict_t		*e;

		// check all the targets
		e = NULL;
		while (1)
		{
			e = G_Find (e, FOFS(targetname), self->target);
			if (!e)
				break;
			if (strcmp(e->classname, "light") != 0)
			{
				if (developer->integer)
				{
					G_Printf ("%s at %s ", self->classname, vtos(self->s.origin));
					G_Printf ("target %s (%s at %s) is not a light\n", self->target, e->classname, vtos(e->s.origin));
				}
			}
			else
			{
				self->enemy = e;
			}
		}

		if (!self->enemy)
		{
			if (developer->integer)
				G_Printf ("%s target %s not found at %s\n", self->classname, self->target, vtos(self->s.origin));
			G_FreeEdict (self);
			return;
		}
	}

	self->timestamp = level.time;
	target_lightramp_think (self);
}

void SP_target_lightramp (edict_t *self)
{
	if (!self->message || strlen(self->message) != 2 || self->message[0] < 'a' || self->message[0] > 'z' || self->message[1] < 'a' || self->message[1] > 'z' || self->message[0] == self->message[1])
	{
		if (developer->integer)
			G_Printf ("target_lightramp has bad ramp (%s) at %s\n", self->message, vtos(self->s.origin));
		G_FreeEdict (self);
		return;
	}
/*
	if (deathmatch->value)
	{
		G_FreeEdict (self);
		return;
	}
*/
	if (!self->target)
	{
		if (developer->integer)
			G_Printf ("%s with no target at %s\n", self->classname, vtos(self->s.origin));
		G_FreeEdict (self);
		return;
	}

	self->r.svflags |= SVF_NOCLIENT;
	self->use = target_lightramp_use;
	self->think = target_lightramp_think;

	self->movedir[0] = self->message[0] - 'a';
	self->movedir[1] = self->message[1] - 'a';
	self->movedir[2] = (self->movedir[1] - self->movedir[0]) / (self->speed / FRAMETIME);
}

//==========================================================

/*QUAKED target_earthquake (1 0 0) (-8 -8 -8) (8 8 8)
When triggered, this initiates a level-wide earthquake.
All players and monsters are affected.
"speed"		severity of the quake (default:200)
"count"		duration of the quake (default:5)
*/

void target_earthquake_think (edict_t *self)
{
	int		i;
	edict_t	*e;

	if (self->last_move_time < level.time)
	{
		G_Sound (self, CHAN_AUTO, self->noise_index, 1.0, ATTN_NONE);
		self->last_move_time = level.time + 0.5;
	}

	for (i=1, e=game.edicts+i; i < game.numentities; i++,e++)
	{
		if (!e->r.inuse)
			continue;
		if (!e->r.client)
			continue;
		if (!e->groundentity)
			continue;

		e->groundentity = NULL;
		e->velocity[0] += crandom()* 150;
		e->velocity[1] += crandom()* 150;
		e->velocity[2] = self->speed * (100.0 / e->mass);
	}

	if (level.time < self->timestamp)
#ifdef SV_FRAMETIME
		self->nextthink = level.timemsec + G_FRAMETIME;
#else
		self->nextthink = level.time + FRAMETIME;
#endif
}

void target_earthquake_use (edict_t *self, edict_t *other, edict_t *activator)
{
	self->timestamp = level.time + self->count;
#ifdef SV_FRAMETIME
	self->nextthink = level.timemsec + G_FRAMETIME;
#else
	self->nextthink = level.time + FRAMETIME;
#endif
	self->activator = activator;
	self->last_move_time = 0;
}

void SP_target_earthquake (edict_t *self)
{
	if (!self->targetname)
		if (developer->integer)
			G_Printf ("untargeted %s at %s\n", self->classname, vtos(self->s.origin));

	if (!self->count)
		self->count = 5;

	if (!self->speed)
		self->speed = 200;

	self->r.svflags |= SVF_NOCLIENT;
	self->think = target_earthquake_think;
	self->use = target_earthquake_use;

	self->noise_index = trap_SoundIndex ("sound/world/quake.wav");
}

//=====================================================

/*QUAKED target_character (0 0 1) ?
used with target_string (must be on same "team")
"count" is position in the string (starts at 1)
*/

void SP_target_character (edict_t *self)
{
	self->movetype = MOVETYPE_PUSH;
	trap_SetBrushModel (self, self->model);
	self->r.solid = SOLID_BSP;
	self->s.frame = 12;
	trap_LinkEntity (self);
	return;
}

//==========================================================

/*QUAKED target_string (0 0 1) (-8 -8 -8) (8 8 8)
*/

void target_string_use (edict_t *self, edict_t *other, edict_t *activator)
{
	edict_t *e;
	int		n, l;
	char	c;

	l = strlen(self->message);
	for (e = self->teammaster; e; e = e->teamchain)
	{
		if (!e->count)
			continue;
		n = e->count - 1;
		if (n > l)
		{
			e->s.frame = 12;
			continue;
		}

		c = self->message[n];
		if (c >= '0' && c <= '9')
			e->s.frame = c - '0';
		else if (c == '-')
			e->s.frame = 10;
		else if (c == ':')
			e->s.frame = 11;
		else
			e->s.frame = 12;
	}
}

void SP_target_string (edict_t *self)
{
	if (!self->message)
		self->message = "";
	self->use = target_string_use;
}

//==========================================================

/*QUAKED target_position (0 .5 0) (-8 -8 -8) (8 8 8)
Aiming target for entities like light, misc_portal_camera and trigger_push (jump pads) in particular.
-------- KEYS --------
targetname : the entity that requires an aiming direction points to this.
notsingle : when set to 1, entity will not spawn in Single Player mode
notfree : when set to 1, entity will not spawn in "Free for all" and "Tournament" modes.
notduel : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
notteam : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. 
notctf : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
-------- NOTES --------
To make a jump pad, place this entity at the highest point of the jump and target it with a trigger_push entity.*/

void SP_target_position (edict_t *self) 
{
	self->r.svflags |= SVF_NOCLIENT;
}


/*QUAKED target_location (0 .5 0) (-8 -8 -8) (8 8 8)
Location marker used by bots and players for team orders and team chat in the course of Teamplay games. The closest target_location in sight is used for the location. If none is in sight, the closest in distance is used.
-------- KEYS --------
message :  name of the location (text string). Displayed in parentheses in front of all team chat and order messages.
count : color of the location text displayed in parentheses during team chat. Set to 0-7 for color.
   0 : white (default)
   1 : red
   2 : green
   3 : yellow
   4 : blue
   5 : cyan
   6 : magenta
   7 : white
notsingle : when set to 1, entity will not spawn in Single Player mode
notfree : when set to 1, entity will not spawn in "Free for all" and "Tournament" modes.
notduel : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
notteam : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. 
notctf : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
*/
void SP_target_location (edict_t *self)
{
	self->r.svflags |= SVF_NOCLIENT;

	if ( self->count ) {
		if ( self->count < 0 ) {
			self->count = 0;
		} else if ( self->count > 7 ) {
			self->count = 7;
		}
	}
}



/*QUAKED target_print (0 .5 0) (-8 -8 -8) (8 8 8) SAMETEAM OTHERTEAM PRIVATE
This will print a message on the center of the screen when triggered. By default, all the clients will see the message.
-------- KEYS --------
message : text string to print on screen.
targetname : the activating trigger points to this.
notsingle : when set to 1, entity will not spawn in Single Player mode
notfree : when set to 1, entity will not spawn in "Free for all" and "Tournament" modes.
notduel : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
notteam : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. 
notctf : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes. (jal: todo)
-------- SPAWNFLAGS --------
SAMETEAM : &1 only players in activator's team will see the message.
OTHERTEAM : &2 only players in other than activator's team will see the message.
PRIVATE : &4 only the player that activates the target will see the message.*/

void SP_target_print_use (edict_t *self, edict_t *other, edict_t *activator)
{
	int n;
	edict_t *player;

	if (activator->r.client && (self->spawnflags & 4)) 
	{
		G_CenterPrintMsg ( activator, self->message );
		return;
	}

	// print to team
	if ( activator->r.client && self->spawnflags & 3 ) {
		edict_t *e;
		for( e = game.edicts + 1; PLAYERNUM(e) < game.maxclients; e++ ) {
			if( e->r.inuse && e->gameTeam ) {
				if( self->spawnflags & 1 && e->gameTeam == activator->gameTeam )
					G_CenterPrintMsg ( e, self->message );
				if( self->spawnflags & 2 && e->gameTeam != activator->gameTeam )
					G_CenterPrintMsg ( e, self->message );
			}
		}
		return;
	}
		
	for (n = 1; n <= game.maxclients; n++)
	{
		player = &game.edicts[n];
		if (!player->r.inuse)
			continue;
		
		G_CenterPrintMsg ( player, self->message );
	}
}

void SP_target_print (edict_t *self)
{
	if ( !self->message ) {
		G_FreeEdict ( self );
		return;
	}

	self->use = SP_target_print_use;
	// do nothing
}


// JALFIXME: We have trigger_relay (and I already commented it should be a target), Q3 has
// this target_relay. IMO we should do the move into target_relay too.

//=============================================================================

/*QUAKED target_relay (0 .7 .7) (-8 -8 -8) (8 8 8) RED_ONLY BLUE_ONLY RANDOM
This can only be activated by other triggers which will cause it in turn to activate its own targets.
-------- KEYS --------
targetname : activating trigger points to this.
target : this points to entities to activate when this entity is triggered.
notfree : when set to 1, entity will not spawn in "Free for all" and "Tournament" modes.
notteam : when set to 1, entity will not spawn in "Teamplay" and "CTF" modes.
notsingle : when set to 1, entity will not spawn in Single Player mode (bot play mode).
-------- SPAWNFLAGS --------
RED_ONLY : only red team players can activate trigger.
BLUE_ONLY : only red team players can activate trigger.
RANDOM : one one of the targeted entities will be triggered at random.*/


