/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "g_local.h"

/*
//
//=================
//check_dodge
//
//This is a support routine used when a client is firing
//a non-instant attack weapon. It checks to see if a
//monster's dodge function should be called.
//=================
//
static void check_dodge (edict_t *self, vec3_t start, vec3_t dir, int speed)
{
	vec3_t	end;
	vec3_t	v;
	trace_t	tr;
	float	eta;

	// easy mode only ducks one quarter the time
	if (skill->integer == 0)
	{
		if (random() > 0.25)
			return;
	}
	VectorMA (start, 8192, dir, end);
	trap_Trace (&tr, start, NULL, NULL, end, self, MASK_SHOT);
	if ((tr.fraction < 1) && (game.edicts[tr.ent].r.svflags & SVF_MONSTER) && (game.edicts[tr.ent].health > 0) && (game.edicts[tr.ent].monsterinfo.dodge) && infront(&game.edicts[tr.ent], self))
	{
		VectorSubtract (tr.endpos, start, v);
		eta = (VectorLength(v) - game.edicts[tr.ent].r.maxs[0]) / speed;
		game.edicts[tr.ent].monsterinfo.dodge (&game.edicts[tr.ent], self, eta);
	}
}


//
//=================
//fire_hit
//
//Used for all impact (hit/punch/slash) attacks
//=================
//
qboolean fire_hit (edict_t *self, vec3_t aim, int damage, int kick)
{
	trace_t		tr;
	vec3_t		forward, right, up;
	vec3_t		v;
	vec3_t		point;
	float		range;
	vec3_t		dir;

	//see if enemy is in range
	VectorSubtract (self->enemy->s.origin, self->s.origin, dir);
	range = VectorLength(dir);
	if (range > aim[0])
		return qfalse;

	if (aim[1] > self->r.mins[0] && aim[1] < self->r.maxs[0])
	{
		// the hit is straight on so back the range up to the edge of their bbox
		range -= self->enemy->r.maxs[0];
	}
	else
	{
		// this is a side hit so adjust the "right" value out to the edge of their bbox
		if (aim[1] < 0)
			aim[1] = self->enemy->r.mins[0];
		else
			aim[1] = self->enemy->r.maxs[0];
	}

	VectorMA (self->s.origin, range, dir, point);

	trap_Trace (&tr, self->s.origin, NULL, NULL, point, self, MASK_SHOT);
	if (tr.fraction < 1)
	{
		if (!game.edicts[tr.ent].takedamage)
			return qfalse;
		// if it will hit any client/monster then hit the one we wanted to hit
		if ((game.edicts[tr.ent].r.svflags & SVF_MONSTER) || (game.edicts[tr.ent].r.client))
			tr.ent = self->enemy - game.edicts;
	}

	AngleVectors(self->s.angles, forward, right, up);
	VectorMA (self->s.origin, range, forward, point);
	VectorMA (point, aim[1], right, point);
	VectorMA (point, aim[2], up, point);
	VectorSubtract (point, self->enemy->s.origin, dir);

	// do the damage
	T_Damage (&game.edicts[tr.ent], self, self, dir, point, vec3_origin, damage, kick/2, DAMAGE_NO_KNOCKBACK, MOD_HIT);

	if (!(game.edicts[tr.ent].r.svflags & SVF_MONSTER) && (!game.edicts[tr.ent].r.client))
		return qfalse;

	// do our special form of knockback here
	VectorMA (self->enemy->r.absmin, 0.5, self->enemy->r.size, v);
	VectorSubtract (v, point, v);
	VectorNormalize (v);
	VectorMA (self->enemy->velocity, kick, v, self->enemy->velocity);
	if (self->enemy->velocity[2] > 0)
		self->enemy->groundentity = NULL;
	return qtrue;
}


//
//=================
//fire_lead
//
//This is an internal support routine used for bullet/pellet based weapons.
//Must match CG_fireLead in cg_events.c
//=================
//
static void fire_lead (edict_t *self, vec3_t start, vec3_t aimdir, vec3_t axis[3], int damage, int kick, int hspread, int vspread, int *seed, int mod)
{
	trace_t		tr;
	vec3_t		dir;
	vec3_t		end;
	float		r;
	float		u;
	vec3_t		water_start;
	int			content_mask = MASK_SHOT | MASK_WATER;

	 trap_Trace (&tr, self->s.origin, NULL, NULL, start, self, MASK_SHOT);
	if (!(tr.fraction < 1.0))
	{
		r = Q_crandom (seed) * hspread;
		u = Q_crandom (seed) * vspread;
		VectorMA (start, 8192, axis[0], end);
		VectorMA (end, r, axis[1], end);
		VectorMA (end, u, axis[2], end);

		if (trap_PointContents (start) & MASK_WATER)
		{
			VectorCopy (start, water_start);
			content_mask &= ~MASK_WATER;
		}

		trap_Trace (&tr, start, NULL, NULL, end, self, content_mask);

		// see if we hit water
		if (tr.contents & MASK_WATER)
		{
			VectorCopy (tr.endpos, water_start);

			if (!VectorCompare (start, tr.endpos))
			{
				vec3_t forward, right, up;

				// change bullet's course when it enters water
				VectorSubtract (end, start, dir);
				VecToAngles (dir, dir);
				AngleVectors (dir, forward, right, up);
				r = Q_crandom (seed) * hspread * 2;
				u = Q_crandom (seed) * vspread * 2;
				VectorMA (water_start, 8192, forward, end);
				VectorMA (end, r, right, end);
				VectorMA (end, u, up, end);
			}

			// re-trace ignoring water this time
			trap_Trace (&tr, water_start, NULL, NULL, end, self, MASK_SHOT);
		}
	}

	// send gun puff / flash
	if (tr.fraction < 1.0)
	{
		if (game.edicts[tr.ent].takedamage)
		{
			T_Damage (&game.edicts[tr.ent], self, self, aimdir, tr.endpos, tr.plane.normal, damage, kick, DAMAGE_BULLET, mod);
		}
		else
		{
			if ( !(tr.surfFlags & SURF_NOIMPACT) )
			{
				if (self->r.client)
					PlayerNoise (self, tr.endpos, PNOISE_IMPACT);
			}
		}
	}
}


//
//=================
//fire_bullet
//
//Fires a single round.  Used for machinegun and chaingun.  Would be fine for
//pistols, rifles, etc....
//=================
//
void fire_bullet (edict_t *self, vec3_t start, vec3_t aimdir, int damage, int kick, int mod)
{
	vec3_t	dir, axis[3];
	edict_t *event;
	int		seed = rand() & 255;

	VecToAngles ( aimdir, dir );
	AngleVectors ( dir, axis[0], axis[1], axis[2] );

	// send the event
	event = G_SpawnEvent ( EV_FIRE_BULLET, seed, start );
	event->s.ownerNum = self - game.edicts;
	event->r.svflags = SVF_FORCEOLDORIGIN;
	VectorScale ( axis[0], 1024, event->s.origin2 );	// DirToByte is too inaccurate

	fire_lead ( self, start, aimdir, axis, damage, kick, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, &seed, mod );
}


//
//=================
//fire_shotgun
//
//Shoots shotgun pellets.  Used by shotgun and super shotgun.
//=================
//
void fire_shotgun (edict_t *self, vec3_t start, vec3_t aimdir, int damage, int kick, int count, int mod)
{
	vec3_t	dir, axis[3];
	edict_t *event;
	int		i, seed = rand() & 255;

	VecToAngles ( aimdir, dir );
	AngleVectors ( dir, axis[0], axis[1], axis[2] );

	// send the event
	event = G_SpawnEvent ( EV_FIRE_SHOTGUN, seed, start );
	event->s.eventCount = count;
	event->s.ownerNum = self - game.edicts;
	event->r.svflags = SVF_FORCEOLDORIGIN;
	VectorScale ( axis[0], 4096, event->s.origin2 );	// DirToByte is too inaccurate

	for ( i = 0; i < count; i++ ) {
		fire_lead ( self, start, aimdir, axis, damage, kick, DEFAULT_SHOTGUN_HSPREAD, DEFAULT_SHOTGUN_VSPREAD, &seed, mod );
	}
}


//
//=================
//fire_blaster
//
//Fires a single blaster bolt.  Used by the blaster and hyper blaster.
//=================
//
void blaster_touch (edict_t *self, edict_t *other, cplane_t *plane, int surfFlags)
{
	if (other == self->r.owner)
		return;
	if (surfFlags & SURF_NOIMPACT)
	{
		G_FreeEdict (self);
		return;
	}

	if (self->r.owner->r.client)
		PlayerNoise(self->r.owner, self->s.origin, PNOISE_IMPACT);

	if (other->takedamage)
	{
		T_Damage (other, self, self->r.owner, self->velocity, self->s.origin, plane->normal, self->dmg, 1, DAMAGE_ENERGY, self->style);
	}
	else
	{
		// turn entity into event
		G_TurnEntityIntoEvent ( self, EV_BLASTER, DirToByte (plane ? plane->normal : NULL) );
		return;
	}

	G_FreeEdict (self);
}

void fire_blaster (edict_t *self, vec3_t start, vec3_t dir, int damage, int speed, int type, int mod)
{
	edict_t	*bolt;
	trace_t	tr;

	VectorNormalize (dir);

	bolt = G_Spawn();
	bolt->r.svflags = SVF_PROJECTILE;
	// when prediction is used against the object
	// (blaster/hyperblaster shots), the player won't be solid clipped against
	// the object. Right now trying to run into a firing hyperblaster
	// is very jerky since you are predicted 'against' the shots.
	VectorCopy (start, bolt->s.origin);
	VectorCopy (start, bolt->s.old_origin);
	VecToAngles (dir, bolt->s.angles);
	VectorScale (dir, speed, bolt->velocity);
	bolt->movetype = MOVETYPE_FLYMISSILE;
	bolt->r.clipmask = MASK_SHOT;
	bolt->r.solid = SOLID_BBOX;
	bolt->s.type = type;
	bolt->s.renderfx |= RF_NOSHADOW;
	VectorClear (bolt->r.mins);
	VectorClear (bolt->r.maxs);
	bolt->s.modelindex = trap_ModelIndex ("models/objects/laser/tris.md2");
	bolt->s.sound = trap_SoundIndex ("sound/misc/lasfly.wav");
	bolt->r.owner = self;
	bolt->touch = blaster_touch;
	bolt->nextthink = level.time + 2;
	bolt->think = G_FreeEdict;
	bolt->dmg = damage;
	bolt->classname = "bolt";
	bolt->style = mod;
	trap_LinkEntity (bolt);

	//newgametypes (this was for q2 monsters, who cares about them)

	trap_Trace (&tr, self->s.origin, NULL, NULL, bolt->s.origin, bolt, MASK_SHOT);
	if (tr.fraction < 1.0)
	{
		VectorMA (bolt->s.origin, -10, dir, bolt->s.origin);
		bolt->touch (bolt, &game.edicts[tr.ent], &tr.plane, 0);
	}
}	
*/

/*
//=================
//fire_grenade
//=================
static void Grenade_Explode (edict_t *ent)
{
	int			mod;

	if (ent->r.owner->r.client)
		PlayerNoise(ent->r.owner, ent->s.origin, PNOISE_IMPACT);

	//FIXME: if we are onground then raise our Z just a bit since we are a point?
	if (ent->enemy)
	{
		float	points;
		vec3_t	v;
		vec3_t	dir;

		VectorAdd (ent->enemy->r.mins, ent->enemy->r.maxs, v);
		VectorMA (ent->enemy->s.origin, 0.5, v, v);
		VectorSubtract (ent->s.origin, v, v);
		points = ent->dmg - 0.5 * VectorLength (v);
		VectorSubtract (ent->enemy->s.origin, ent->s.origin, dir);
		if (ent->spawnflags & 1)
			mod = MOD_HANDGRENADE;
		else
			mod = MOD_GRENADE;
		T_Damage (ent->enemy, ent, ent->r.owner, dir, ent->s.origin, vec3_origin, (int)points, (int)points, DAMAGE_RADIUS, mod);
	}

	if (ent->spawnflags & 2)
		mod = MOD_HELD_GRENADE;
	else if (ent->spawnflags & 1)
		mod = MOD_HG_SPLASH;
	else
		mod = MOD_G_SPLASH;
	T_RadiusDamage(ent, ent->r.owner, ent->dmg, ent->radius_dmg, ent->enemy, ent->dmg_radius, mod);

	// turn entity into event
	VectorMA (ent->s.origin, -0.02, ent->velocity, ent->s.origin);
	G_TurnEntityIntoEvent ( ent, ent->groundentity ? EV_GRENADE_EXPLOSION : EV_ROCKET_EXPLOSION, DirToByte (ent->movedir) );
}

static void Grenade_Touch (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	if (other == ent->r.owner)
		return;
	if (surfFlags & SURF_NOIMPACT)
	{
		G_FreeEdict (ent);
		return;
	}

	if (!other->takedamage)
	{
		G_AddEvent (ent, EV_GRENADE_BOUNCE, 0, qtrue);
		return;
	}

	ent->enemy = other;
	Grenade_Explode (ent);
}

static void LaserGrenade_SetupLaser (edict_t *ent)
{
	edict_t	*owner;
	edict_t *laser;

	if( !(owner = ent->r.owner) ) {
		ent->nextthink = level.time + FRAMETIME;
		ent->think = G_FreeEdict;
		return;
	}

	laser = G_Spawn();
	laser->s.sound = trap_SoundIndex ("world/laser.wav");
	laser->classname = "grenade_laser";
	laser->enemy = NULL;
	laser->target = NULL;
  	laser->r.owner = ent;
	laser->activator = owner;
  	laser->dmg = 25;
	laser->delay = 10;
	laser->count = MOD_LASER_TRAP;
	laser->spawnflags = 1;

//newgametypes[start] JALFIXME: just hacked it in being always color red. The color should match the team color
	//if( !ctf->integer || !owner->r.client || owner->r.client->resp.ctf_team == CTF_TEAM1 )
		laser->spawnflags |= 2;		// red
	//else		
		//laser->spawnflags |= 8;		// blue
//newgametypes[end]

	VectorCopy( ent->s.origin, laser->s.origin );
	VectorCopy( ent->s.angles, laser->s.angles );

	target_laser_start (laser);
}

static void LaserGrenade_Use (edict_t *self, edict_t *other, edict_t *activator)
{
	self->enemy = NULL;
	Grenade_Explode (self);
}

static void LaserGrenade_Touch (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	if (other == ent->r.owner)
		return;
	if (surfFlags & SURF_NOIMPACT)
	{
		G_FreeEdict (ent);
		return;
	}

	if (other != world)
	{
		if (other->takedamage)
			ent->enemy = other;
		Grenade_Explode (ent);
		return;
	}

	ent->use = LaserGrenade_Use;
	ent->touch = NULL;
	ent->movetype = MOVETYPE_NONE;
	ent->s.sound = 0;
	G_AddEvent (ent, EV_GRENADE_BOUNCE, 0, qtrue);

	ent->nextthink = level.time + 2;
	ent->think = LaserGrenade_SetupLaser;

	VectorClear (ent->velocity);
	VectorClear (ent->avelocity);

	VecToAngles (plane->normal, ent->s.angles);
	VectorCopy (plane->normal, ent->movedir);

	trap_LinkEntity (ent);
}

void fire_grenade (edict_t *self, vec3_t start, vec3_t aimdir, int damage, int speed, float timer, float damage_radius)
{
	edict_t	*grenade;
	vec3_t	dir;
	vec3_t	forward, right, up;

	VecToAngles (aimdir, dir);
	AngleVectors (dir, forward, right, up);

	grenade = G_Spawn();
	VectorCopy (start, grenade->s.origin);
	VectorScale (aimdir, speed, grenade->velocity);
	grenade->velocity[2] += 200 * up[2] + crandom() * 10.0;
	VectorMA (grenade->velocity, crandom() * 10.0, right, grenade->velocity);
	VectorSet (grenade->avelocity, 300, 300, 300);
	VectorSet (grenade->movedir, 0, 0, 1);
	grenade->movetype = MOVETYPE_BOUNCEGRENADE;
	grenade->r.clipmask = MASK_SHOT;
	grenade->r.solid = SOLID_BBOX;
	grenade->s.type = ET_GRENADE;
	grenade->s.renderfx |= RF_NOSHADOW;
	VectorClear (grenade->r.mins);
	VectorClear (grenade->r.maxs);
	grenade->s.modelindex = trap_ModelIndex ("models/ammo/grenade1.md3");
	grenade->r.owner = self;
	grenade->touch = Grenade_Touch;
	grenade->nextthink = level.time + timer;
	grenade->think = Grenade_Explode;
	grenade->use = NULL;
	grenade->dmg = damage;
	grenade->dmg_radius = damage_radius;
	grenade->classname = "grenade";

	trap_LinkEntity (grenade);
}


void fire_grenade2 (edict_t *self, vec3_t start, vec3_t aimdir, int damage, int speed, float timer, float damage_radius, qboolean held)
{
	edict_t	*grenade;
	vec3_t	dir;
	vec3_t	forward, right, up;

	VecToAngles (aimdir, dir);
	AngleVectors (dir, forward, right, up);

	grenade = G_Spawn();
	VectorCopy (start, grenade->s.origin);
	VectorScale (aimdir, speed, grenade->velocity);
	VectorMA (grenade->velocity, 200 + crandom() * 10.0, up, grenade->velocity);
	VectorMA (grenade->velocity, crandom() * 10.0, right, grenade->velocity);
	VectorSet (grenade->avelocity, 300, 300, 300);
	VectorSet (grenade->movedir, 0, 0, 1);
	grenade->movetype = MOVETYPE_BOUNCE;
	grenade->r.clipmask = MASK_SHOT;
	grenade->r.solid = SOLID_BBOX;
	grenade->s.renderfx |= RF_NOSHADOW;
	VectorClear (grenade->r.mins);
	VectorClear (grenade->r.maxs);
	grenade->s.modelindex = trap_ModelIndex ("models/objects/grenade2/tris.md2");
	grenade->r.owner = self;
	grenade->touch = LaserGrenade_Touch;
	grenade->use = NULL;
	grenade->nextthink = level.time + timer;
	grenade->think = Grenade_Explode;
	grenade->dmg = damage;
	grenade->dmg_radius = damage_radius;
	grenade->classname = "hgrenade";
	if (held)
		grenade->spawnflags = 3;
	else
		grenade->spawnflags = 1;
	grenade->s.sound = trap_SoundIndex("sound/weapons/hgrenc1b.wav");

	if (timer <= 0.0)
		Grenade_Explode (grenade);
	else
	{
		//splitmodels (handgrenade drop sound moved to cgame)
		trap_LinkEntity (grenade);
	}
}


//
//=================
//fire_rocket
//=================
//
void rocket_touch (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	if (other == ent->r.owner)
		return;
	if (surfFlags & SURF_NOIMPACT)
	{
		G_FreeEdict (ent);
		return;
	}

	if (ent->r.owner->r.client)
		PlayerNoise (ent->r.owner, ent->s.origin, PNOISE_IMPACT);

	if (other->takedamage)
		T_Damage (other, ent, ent->r.owner, ent->velocity, ent->s.origin, plane->normal, ent->dmg, 0, 0, MOD_ROCKET);

	T_RadiusDamage(ent, ent->r.owner, ent->dmg, ent->radius_dmg, other, ent->dmg_radius, MOD_R_SPLASH);

	// turn entity into event
	VectorMA ( ent->s.origin, -0.02, ent->velocity, ent->s.origin );
	G_TurnEntityIntoEvent ( ent, EV_ROCKET_EXPLOSION, DirToByte (plane ? plane->normal : NULL) );
}

void fire_rocket (edict_t *self, vec3_t start, vec3_t dir, int damage, int speed, float damage_radius, int radius_damage)
{
	edict_t	*rocket;

	rocket = G_Spawn();
	VectorCopy (start, rocket->s.origin);
	VectorCopy (dir, rocket->movedir);
	VecToAngles (dir, rocket->s.angles);
	VectorScale (dir, speed, rocket->velocity);
	rocket->movetype = MOVETYPE_FLYMISSILE;
	rocket->r.clipmask = MASK_SHOT;
	rocket->r.solid = SOLID_BBOX;
	rocket->s.type = ET_ROCKET;
	rocket->s.renderfx |= RF_NOSHADOW;
	VectorClear (rocket->r.mins);
	VectorClear (rocket->r.maxs);
	rocket->s.modelindex = trap_ModelIndex ("models/ammo/rocket/rocket.md3");
	rocket->r.owner = self;
	rocket->touch = rocket_touch;
	rocket->nextthink = level.time + 8000/speed;
	rocket->think = G_FreeEdict;
	rocket->dmg = damage;
	rocket->radius_dmg = radius_damage;
	rocket->dmg_radius = damage_radius;
	rocket->s.sound = trap_SoundIndex ("sound/weapons/rocket/rockfly.wav");
	rocket->classname = "rocket";

	//newgametypes (this was for q2 monsters, who cares about them)

	trap_LinkEntity (rocket);
}


//
//=================
//fire_rail
//=================
//
void fire_rail (edict_t *self, vec3_t start, vec3_t aimdir, int damage, int kick)
{
	vec3_t		from;
	vec3_t		end;
	trace_t		tr;
	edict_t		*ignore, *event;
	int			mask;

	VectorMA (start, 8192, aimdir, end);
	VectorCopy (start, from);
	ignore = self;
	mask = MASK_SHOT;
	while (ignore)
	{
		trap_Trace (&tr, from, NULL, NULL, end, ignore, mask);

		//ZOID--added so rail goes through SOLID_BBOX entities (gibs, etc)
		if ((game.edicts[tr.ent].r.svflags & SVF_MONSTER) || (game.edicts[tr.ent].r.client) ||
			(game.edicts[tr.ent].r.solid == SOLID_BBOX))
			ignore = &game.edicts[tr.ent];
		else
			ignore = NULL;

		if ((&game.edicts[tr.ent] != self) && (game.edicts[tr.ent].takedamage))
			T_Damage (&game.edicts[tr.ent], self, self, aimdir, tr.endpos, tr.plane.normal, damage, kick, 0, MOD_RAILGUN);

		VectorCopy (tr.endpos, from);
	}

	// send gun puff / flash
	event = G_SpawnEvent ( EV_RAILTRAIL, 0, start );
	event->r.svflags = SVF_FORCEOLDORIGIN;
	VectorCopy ( tr.endpos, event->s.origin2 );
	event->s.ownerNum = self->s.number;//splitmodels (identify who fired for cgame)

	if (self->r.client)
		PlayerNoise(self, tr.endpos, PNOISE_IMPACT);
}
*/

/*
//
//=================
//fire_bfg
//=================
//
void bfg_explode (edict_t *self)
{
	edict_t	*ent;
	float	points;
	vec3_t	v;
	float	dist;

	if (self->s.frame == 0)
	{
		// the BFG effect
		ent = NULL;
		while ((ent = findradius(ent, self->s.origin, self->dmg_radius)) != NULL)
		{
			if (!ent->takedamage)
				continue;
			if (ent == self->r.owner)
				continue;
			if (!CanDamage (ent, self))
				continue;
			if (!CanDamage (ent, self->r.owner))
				continue;

			VectorAdd (ent->r.mins, ent->r.maxs, v);
			VectorMA (ent->s.origin, 0.5, v, v);
			VectorSubtract (self->s.origin, v, v);
			dist = VectorLength(v);
			points = self->radius_dmg * (1.0 - sqrt(dist/self->dmg_radius));
			if (ent == self->r.owner)
				points = points * 0.5;

			G_SpawnEvent ( EV_BFG_EXPLOSION, 0, ent->s.origin );
			T_Damage (ent, self, self->r.owner, self->velocity, ent->s.origin, vec3_origin, (int)points, 0, DAMAGE_ENERGY, MOD_BFG_EFFECT);
		}
	}

	self->nextthink = level.time + FRAMETIME;
	self->count++;
	if (self->count == 5)
		self->think = G_FreeEdict;
}

void bfg_touch (edict_t *self, edict_t *other, cplane_t *plane, int surfFlags)
{
	if (other == self->r.owner)
		return;
	if (surfFlags & SURF_NOIMPACT)
	{
		G_FreeEdict (self);
		return;
	}

	if (self->r.owner->r.client)
		PlayerNoise(self->r.owner, self->s.origin, PNOISE_IMPACT);

	// core explosion - prevents firing it into the wall/floor
	if (other->takedamage)
		T_Damage (other, self, self->r.owner, self->velocity, self->s.origin, plane->normal, 200, 0, 0, MOD_BFG_BLAST);
	T_RadiusDamage(self, self->r.owner, 200, 0, other, 100, MOD_BFG_BLAST);

	G_Sound (self, CHAN_VOICE, trap_SoundIndex ("sound/weapons/bfg__x1b.wav"), 1, ATTN_NORM);
	self->r.solid = SOLID_NOT;
	self->touch = NULL;
	self->r.svflags = SVF_NOCLIENT;
	VectorMA (self->s.origin, -1 * FRAMETIME, self->velocity, self->s.origin);
	VectorClear (self->velocity);
	self->think = bfg_explode;
	self->nextthink = level.time + FRAMETIME;
	self->enemy = other;
	self->count = 0;
	trap_LinkEntity (self);

	G_SpawnEvent ( EV_BFG_BIGEXPLOSION, 0, self->s.origin );
}


void bfg_think (edict_t *self)
{
	edict_t	*ent;
	edict_t	*ignore;
	edict_t *event;
	vec3_t	point;
	vec3_t	dir;
	vec3_t	start;
	vec3_t	end;
	int		dmg;
	trace_t	tr;

	if (game.gametype != GAMETYPE_SP 
	//&& game.gametype != GAMETYPE_COOP
	)//newgametypes (so idiot)
		dmg = 5;
	else
		dmg = 10;

	ent = NULL;
	while ((ent = findradius(ent, self->s.origin, 256)) != NULL)
	{
		if (ent == self)
			continue;
		if (ent == self->r.owner)
			continue;
		if (!ent->takedamage)
			continue;
		if (!(ent->r.svflags & SVF_MONSTER) && !(ent->r.client) && (strcmp(ent->classname, "misc_explobox") != 0))
			continue;

//ZOID
//newgametypes
//ZOID

		VectorMA (ent->r.absmin, 0.5, ent->r.size, point);

		VectorSubtract (point, self->s.origin, dir);
		VectorNormalize (dir);

		ignore = self;
		VectorCopy (self->s.origin, start);
		VectorMA (start, 2048, dir, end);
		while(1)
		{
			trap_Trace (&tr, start, NULL, NULL, end, ignore, MASK_SHOT);

			if (!tr.ent)
				break;

			// hurt it if we can
			if ((game.edicts[tr.ent].takedamage) && !(game.edicts[tr.ent].flags & FL_IMMUNE_LASER) && (&game.edicts[tr.ent] != self->r.owner))
				T_Damage (&game.edicts[tr.ent], self, self->r.owner, dir, tr.endpos, vec3_origin, dmg, 1, DAMAGE_ENERGY, MOD_BFG_LASER);

			// if we hit something that's not a monster or player we're done
			if (!(game.edicts[tr.ent].r.svflags & SVF_MONSTER) && (!game.edicts[tr.ent].r.client))
			{
				edict_t *event;

				event = G_SpawnEvent ( EV_LASER_SPARKS, DirToByte (tr.plane.normal), tr.endpos );
				event->s.eventCount = 4;
				event->s.skinnum = self->s.skinnum;
				break;
			}

			ignore = &game.edicts[tr.ent];
			VectorCopy (tr.endpos, start);
		}

		event = G_SpawnEvent ( EV_BFG_LASER, 0, self->s.origin );
		event->r.svflags = SVF_FORCEOLDORIGIN;
		VectorCopy ( tr.endpos, event->s.origin2 );
	}

	self->nextthink = level.time + FRAMETIME;
}


void fire_bfg (edict_t *self, vec3_t start, vec3_t dir, int damage, int speed, float damage_radius)
{
	edict_t	*bfg;

	bfg = G_Spawn();
	VectorCopy (start, bfg->s.origin);
	VectorCopy (dir, bfg->movedir);
	VectorClear (bfg->s.angles);
	VectorScale (dir, speed, bfg->velocity);
	bfg->movetype = MOVETYPE_FLYMISSILE;
	bfg->r.clipmask = MASK_SHOT;
	bfg->r.solid = SOLID_BBOX;
	bfg->s.type = ET_BFG;
	VectorClear (bfg->r.mins);
	VectorClear (bfg->r.maxs);
	bfg->s.modelindex = trap_ModelIndex ("sprites/s_bfg1.sp2");
	bfg->r.owner = self;
	bfg->touch = bfg_touch;
	bfg->nextthink = level.time + 8000/speed;
	bfg->think = G_FreeEdict;
	bfg->radius_dmg = damage;
	bfg->dmg_radius = damage_radius;
	bfg->classname = "bfg blast";
	bfg->s.sound = trap_SoundIndex ("sound/weapons/bfg__l1a.wav");

	bfg->think = bfg_think;
	bfg->nextthink = level.time + FRAMETIME;
	bfg->teammaster = bfg;
	bfg->teamchain = NULL;

	//newgametypes (this was for q2 monsters, who cares about them)

	trap_LinkEntity (bfg);
}
*/

//==================
//W_Fire_Lead
//the seed is important to be as pointer for cgame prediction accuracy
//==================
static void W_Fire_Lead (edict_t *self, vec3_t start, vec3_t aimdir, vec3_t axis[3], int damage, int kick, int hspread, int vspread, int *seed, int dflags, int mod)
{
	trace_t		tr;
	vec3_t		dir;
	vec3_t		end;
	float		r;
	float		u;
	vec3_t		water_start;
	int			content_mask = MASK_SHOT | MASK_WATER;

	 trap_Trace (&tr, self->s.origin, NULL, NULL, start, self, MASK_SHOT);
	if (!(tr.fraction < 1.0))
	{
		r = Q_crandom (seed) * hspread;
		u = Q_crandom (seed) * vspread;
		VectorMA (start, 8192, axis[0], end);
		VectorMA (end, r, axis[1], end);
		VectorMA (end, u, axis[2], end);

		if (trap_PointContents (start) & MASK_WATER)
		{
			VectorCopy (start, water_start);
			content_mask &= ~MASK_WATER;
		}

		trap_Trace (&tr, start, NULL, NULL, end, self, content_mask);

		// see if we hit water
		if (tr.contents & MASK_WATER)
		{
			VectorCopy (tr.endpos, water_start);

			if (!VectorCompare (start, tr.endpos))
			{
				vec3_t forward, right, up;

				// change bullet's course when it enters water
				VectorSubtract (end, start, dir);
				VecToAngles (dir, dir);
				AngleVectors (dir, forward, right, up);
				r = Q_crandom (seed) * hspread * 2;
				u = Q_crandom (seed) * vspread * 2;
				VectorMA (water_start, 8192, forward, end);
				VectorMA (end, r, right, end);
				VectorMA (end, u, up, end);
			}

			// re-trace ignoring water this time
			trap_Trace (&tr, water_start, NULL, NULL, end, self, MASK_SHOT);
		}
	}

	// send gun puff / flash
	if (tr.fraction < 1.0)
	{
		if (game.edicts[tr.ent].takedamage)
		{
			T_Damage (&game.edicts[tr.ent], self, self, aimdir, tr.endpos, tr.plane.normal, damage, kick, dflags, mod);
		}
		else
		{
			if ( !(tr.surfFlags & SURF_NOIMPACT) )
			{
			}
		}
	}
}


//==================
//W_Touch_Projectile - Generic projectile touch func. Only for replacement in tests
//==================
void W_Touch_Projectile (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	vec3_t dir;

	//don't hurt onwner for the 1st second?
	if (other == ent->r.owner) {
		if( !g_projectile_touch_owner->integer ||
			(g_projectile_touch_owner->integer && ent->timestamp + 1.0 > level.time) )
			return;
	}

	if (surfFlags & SURF_NOIMPACT)
	{
		G_FreeEdict (ent);
		return;
	}

	if (other->takedamage)
	{
		VectorSubtract( other->s.origin, ent->s.origin, dir );
		VectorNormalize( dir );
		T_Damage (other, ent, ent->r.owner, dir, ent->s.origin, plane->normal, ent->dmg, ent->dmg, 0, MOD_EXPLOSIVE);
	}
	T_RadiusDamage(ent, ent->r.owner, ent->dmg, ent->dmg_knockback, ent->radius_dmg, other, ent->dmg_radius, MOD_EXPLOSIVE);
	
	// turn entity into event
	VectorMA ( ent->s.origin, -0.02, ent->velocity, ent->s.origin );
	G_TurnEntityIntoEvent ( ent, EV_EXPLOSION1, DirToByte (plane ? plane->normal : NULL) );
	ent->s.skinnum = FIRE_MODE_STRONG;
}


//==================
//W_Fire_Projectile - Spawn a generic projectile without a model, touch func, sound nor mod
//==================
edict_t *W_Fire_Projectile( edict_t *self, vec3_t start, vec3_t dir, int speed,
					   int damage, int knockback, int radius_damage, int radius, int timeout )
{
	edict_t	*projectile;

	VectorNormalize (dir);

	projectile = G_Spawn();
	VectorCopy (start, projectile->s.origin);
	VectorCopy (start, projectile->s.old_origin);
	VecToAngles (dir, projectile->s.angles);
	VectorScale (dir, speed, projectile->velocity);
	projectile->movetype = MOVETYPE_FLYMISSILE;
	projectile->r.clipmask = MASK_SHOT;
	projectile->r.solid = SOLID_BBOX;
	projectile->s.renderfx = RF_NOSHADOW;
	VectorClear (projectile->r.mins);
	VectorClear (projectile->r.maxs);
	//projectile->s.modelindex = trap_ModelIndex ("models/objects/projectile/plasmagun/proj_plasmagun2.md3");
	projectile->s.modelindex = 0;
	projectile->r.owner = self;
	projectile->touch = W_Touch_Projectile; //generic one. Should be replaced after calling this func
#ifdef SV_FRAMETIME
	projectile->nextthink = level.timemsec + timeout;
#else
	projectile->nextthink = level.time + timeout * 0.001;
#endif
	projectile->think = G_FreeEdict;
	projectile->dmg = damage;
	projectile->dmg_knockback = knockback;
	projectile->radius_dmg = radius_damage;
	projectile->dmg_radius = radius;
	projectile->classname = "projectile"; //generic one. Should be replaced after calling this func.
	projectile->style = 0;
	projectile->s.sound = 0;
	projectile->timestamp = level.time;

	trap_LinkEntity (projectile);

	return projectile;
}


//	------------ the actual weapons --------------


//==================
//W_Fire_Blade
//==================
void W_Fire_Blade ( edict_t *self, int range, vec3_t start, vec3_t aimdir, int damage, int knockback, int mod)
{
	edict_t *event, *other = NULL;
	vec3_t	end;
	trace_t	trace;

	VectorMA( start, range, aimdir, end );
	trap_Trace (&trace, start, NULL, NULL, end, self, MASK_SHOT);
	if( trace.fraction == 1.0 ) //didn't touch anything
		return;

	// find out what touched
	other = &game.edicts[trace.ent];
	if( !other->takedamage ) // it was the world
	{
		// wall impact
		VectorMA ( trace.endpos, -0.02, aimdir, end );
		event = G_SpawnEvent ( EV_BLADE_IMPACT, 0, end );
		event->s.ownerNum = self - game.edicts;
		VectorCopy( trace.plane.normal, event->s.origin2);
		event->r.svflags = SVF_FORCEOLDORIGIN;
		return;
	}

	// it was a player
	T_Damage (other, self, self, aimdir, other->s.origin, vec3_origin, damage, knockback, 0, mod);
}

//==================
//W_Touch_GunbladeBlast
//==================
void W_Touch_GunbladeBlast (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	vec3_t	dir;
	int		power;

	//don't hurt owner for the 1st second
	if (other == ent->r.owner) {
		if( !g_projectile_touch_owner->integer ||
			(g_projectile_touch_owner->integer && ent->timestamp + 1.0 > level.time) )
			return;
	}

	//if (surfFlags & SURF_NOIMPACT)
	//{
	//	G_FreeEdict (ent);
	//	return;
	//}

	if (other->takedamage)
	{
		VectorSubtract( other->s.origin, ent->s.origin, dir );
		VectorNormalize( dir );
		T_Damage (other, ent, ent->r.owner, dir, ent->s.origin, plane->normal, ent->dmg, ent->dmg_knockback, 0, ent->style);
	}
	T_RadiusDamage(ent, ent->r.owner, ent->dmg, ent->dmg_knockback, ent->radius_dmg, other, ent->dmg_radius, MOD_GUNBLADE_S); //fixme : splash mod
	
	// add explosion event
	if( !(surfFlags & SURF_NOIMPACT) && !other->takedamage )
	{
		edict_t *event;

		power = (int)(128.0f * ((float)ent->dmg / (float)g_weaponInfos[WEAP_GUNBLADE].firedef->damage) );
		if( power > 128 ) power = 128;
		
		event = G_SpawnEvent( EV_GUNBLADEBLAST_IMPACT, DirToByte (plane ? plane->normal : NULL), ent->s.origin );
		event->s.skinnum = power;
	}

	// free at next frame
	ent->think = G_FreeEdict;
	ent->touch = NULL;
#ifdef SV_FRAMETIME
	ent->nextthink = level.timemsec + G_FRAMETIME;
#else
	ent->nextthink = level.time + FRAMETIME;
#endif
}

//==================
//W_Fire_GunbladeBlast
//==================
void W_Fire_GunbladeBlast ( edict_t *self, vec3_t start, vec3_t dir, int damage, int knockback, int radius_damage, int radius, int speed, int timeout, int mod )
{
	edict_t	*blast;

	blast = W_Fire_Projectile(self, start, dir, speed, damage, knockback, radius_damage, radius, timeout);
	blast->s.modelindex = trap_ModelIndex ("models/objects/projectile/gunblade/proj_gunbladeblast.md3");
	blast->s.type = ET_BLASTER;
	blast->touch = W_Touch_GunbladeBlast;
	blast->classname = "Gunblade blast";
	blast->s.renderfx |= RF_FULLBRIGHT;
	blast->style = mod;

	blast->s.sound = trap_SoundIndex ( S_WEAPON_PLASMAGUN_S_FLY );
}



//==================
//W_Fire_Gunblade_Bullet
//==================
void W_Fire_Gunblade_Bullet ( edict_t *self, vec3_t start, vec3_t aimdir, int damage, int knockback, int mod)
{
	vec3_t	dir, axis[3];
	edict_t *event;
	int		seed = rand() & 255;
	//int		kick = 4;

	VecToAngles ( aimdir, dir );
	AngleVectors ( dir, axis[0], axis[1], axis[2] );

	// send the event
	event = G_SpawnEvent ( EV_FIRE_BULLET, seed, start );
	//event->s.eventCount = 1;
	event->s.ownerNum = self - game.edicts;
	event->r.svflags = SVF_FORCEOLDORIGIN;
	VectorScale ( axis[0], 1024, event->s.origin2 );	// DirToByte is too inaccurate

	W_Fire_Lead (self, start, aimdir, axis, damage, knockback, DEFAULT_BULLET_HSPREAD, DEFAULT_BULLET_VSPREAD, &seed, DAMAGE_BULLET, mod);
}

//==================
//W_Fire_Riotgun
//==================
void W_Fire_Riotgun (edict_t *self, vec3_t start, vec3_t aimdir, int damage, int knockback, int hspread, int vspread, int count, int dflags, int mod)
{
	vec3_t	dir, axis[3];
	edict_t *event;
	int		i, seed = rand() & 255;

	VecToAngles ( aimdir, dir );
	AngleVectors ( dir, axis[0], axis[1], axis[2] );
	
	// send the event
	event = G_SpawnEvent ( EV_FIRE_RIOTGUN, seed, start );
	event->s.eventCount = count;
	event->s.ownerNum = self - game.edicts;
	event->r.svflags = SVF_FORCEOLDORIGIN;
	VectorScale ( axis[0], 4096, event->s.origin2 );	// DirToByte is too inaccurate

	//send spreads inside these (the free shorts I found)
	event->s.light = hspread;
	event->s.skinnum = vspread;

	for (i = 0; i < count; i++)
		W_Fire_Lead (self, start, aimdir, axis, damage, knockback, hspread, vspread, &seed, dflags, mod);
}

//==================
//W_Grenade_Explode
//==================
static void W_Grenade_Explode (edict_t *ent)
{
	vec3_t		origin;

	//FIXME: if we are onground then raise our Z just a bit since we are a point?
	if (ent->enemy)
	{
//		float	points;
//		vec3_t	v;
		vec3_t	dir;

/*		VectorAdd (ent->enemy->r.mins, ent->enemy->r.maxs, v);
		VectorMA (ent->enemy->s.origin, 0.5, v, v);
		VectorSubtract (ent->s.origin, v, v);
		points = ent->dmg;// - 0.5 * VectorLength (v);
*/
		VectorSubtract (ent->enemy->s.origin, ent->s.origin, dir);
		T_Damage (ent->enemy, ent, ent->r.owner, dir, ent->s.origin, vec3_origin, ent->dmg, ent->dmg_knockback, 0, ent->style);
	}

	T_RadiusDamage(ent,
		ent->r.owner,
		ent->dmg,
		ent->dmg_knockback,
		ent->radius_dmg,
		ent->enemy,
		ent->dmg_radius,
		(ent->style == MOD_GRENADE_S) ? MOD_GRENADE_SPLASH_S : MOD_GRENADE_SPLASH_W 
		);

	VectorMA (ent->s.origin, -0.02, ent->velocity, origin);
	G_TurnEntityIntoEvent ( ent, EV_GRENADE_EXPLOSION, ent->groundentity ? DirToByte (ent->movedir) : 0 );
	ent->s.skinnum = (ent->style == MOD_GRENADE_S) ? FIRE_MODE_STRONG : FIRE_MODE_WEAK ;
}

//==================
//W_Touch_Grenade
//==================
static void W_Touch_Grenade (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	//don't hurt owner for the first second
	if (other == ent->r.owner) {
		if( !g_projectile_touch_owner->integer ||
			(g_projectile_touch_owner->integer && ent->timestamp + 1.0 > level.time) )
			return;
	}

	if ( surfFlags & SURF_NOIMPACT )
	{
		G_FreeEdict (ent);
		return;
	}

	if (!other->takedamage)
	{
		if( ent->style == MOD_GRENADE_S )
			G_AddEvent (ent, EV_GRENADE_BOUNCE, FIRE_MODE_STRONG, qtrue);
		else
			G_AddEvent (ent, EV_GRENADE_BOUNCE, FIRE_MODE_WEAK, qtrue);

		return;
	}

	ent->enemy = other;
	W_Grenade_Explode (ent);
}

//==================
//W_Fire_Grenade
//==================
void W_Fire_Grenade (edict_t *self, vec3_t start, vec3_t aimdir, int speed, int damage, int knockback, int radius_damage, float damage_radius, int timeout, int mod )
{
	edict_t	*grenade;
	vec3_t	dir;
	vec3_t	forward, right, up;

	VecToAngles (aimdir, dir);
	AngleVectors (dir, forward, right, up);

	grenade = G_Spawn();

	VectorCopy (start, grenade->s.origin);
	VectorScale (aimdir, speed, grenade->velocity);
	VectorMA (grenade->velocity, 200 + crandom() * 10.0, up, grenade->velocity);
	VectorMA (grenade->velocity, crandom() * 10.0, right, grenade->velocity);
	VectorSet (grenade->avelocity, 300, 300, 300);
	VectorSet (grenade->movedir, 0, 0, 1);
	grenade->movetype = MOVETYPE_BOUNCEGRENADE;
	grenade->r.clipmask = MASK_SHOT;
	grenade->r.solid = SOLID_BBOX;
	//grenade->s.effects |= EF_GRENADE;
	grenade->s.renderfx |= RF_NOSHADOW;
	VectorClear (grenade->r.mins);
	VectorClear (grenade->r.maxs);
	if ( mod == MOD_GRENADE_S )	grenade->s.modelindex = trap_ModelIndex (PATH_GRENADE_STRONG);
	else						grenade->s.modelindex = trap_ModelIndex (PATH_GRENADE_WEAK);
	grenade->r.owner = self;
	grenade->touch = W_Touch_Grenade;
	grenade->use = NULL;
#ifdef SV_FRAMETIME
	grenade->nextthink = level.timemsec + timeout;
#else
	grenade->nextthink = level.time + timeout * 0.001;
#endif
	grenade->think = W_Grenade_Explode;
	grenade->dmg = damage;
	grenade->dmg_knockback = knockback;
	grenade->dmg_radius = damage_radius;
	grenade->radius_dmg = radius_damage;
	grenade->style = mod;
	grenade->classname = "grenade";
	grenade->timestamp = level.time;

	//grenade->s.sound = trap_SoundIndex ( S_WEAPON_GRENADE_FLY );

	trap_LinkEntity (grenade);
}

//==================
//W_Touch_Rocket
//==================
void W_Touch_Rocket (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	int mod_splash, fire_mode;
	vec3_t dir;

	//Com_Printf( "%f\n", VectorLength( ent->velocity ) );
	//don't hurt owner for the 1st second
	if (other == ent->r.owner) {
		if( !g_projectile_touch_owner->integer ||
			(g_projectile_touch_owner->integer && ent->timestamp + 1.0 > level.time) )
			return;
	}

	if ( ent->style == MOD_ROCKET_S)
	{
		fire_mode = FIRE_MODE_STRONG;
		mod_splash = MOD_ROCKET_SPLASH_S;
	}
	else
	{
		//ent->style = MOD_ROCKET_W;
		fire_mode = FIRE_MODE_WEAK;
		mod_splash = MOD_ROCKET_SPLASH_W;
	}

	if (other->takedamage)
	{
		VectorSubtract( other->s.origin, ent->s.origin, dir );
		VectorNormalize( dir );
		T_Damage (other, ent, ent->r.owner, dir, ent->s.origin, plane->normal, ent->dmg, ent->dmg_knockback, 0, ent->style);
	}
	T_RadiusDamage(ent, ent->r.owner, ent->dmg, ent->dmg_knockback, ent->radius_dmg, other, ent->dmg_radius, mod_splash);
	
	// spawn the explosion
	if( !(surfFlags & SURF_NOIMPACT) )
	{
		edict_t *event;
		vec3_t	explosion_origin;
		
		VectorMA ( ent->s.origin, -0.02, ent->velocity, explosion_origin );
		event = G_SpawnEvent( EV_ROCKET_EXPLOSION, DirToByte (plane ? plane->normal : NULL), explosion_origin );
		event->s.skinnum = fire_mode;
	}

	// free the rocket at next frame
	ent->think = G_FreeEdict;
	ent->touch = NULL;
#ifdef SV_FRAMETIME
	ent->nextthink = level.timemsec + G_FRAMETIME;
#else
	ent->nextthink = level.time + FRAMETIME;
#endif
}

//==================
//W_Fire_Rocket
//==================
void W_Fire_Rocket (edict_t *self, vec3_t start, vec3_t dir, int speed, int damage, int knockback, int radius_damage, int radius, int timeout, int mod )
{
	edict_t	*rocket;

	rocket = W_Fire_Projectile(self, start, dir, speed, damage, knockback, radius_damage, radius, timeout);

	rocket->s.type = ET_ROCKET; //rocket trail sfx
	if ( mod == MOD_ROCKET_S )	rocket->s.modelindex = trap_ModelIndex (PATH_ROCKET_STRONG);
	else						rocket->s.modelindex = trap_ModelIndex (PATH_ROCKET_WEAK);
	rocket->touch = W_Touch_Rocket;
	rocket->think = G_FreeEdict;
	rocket->s.renderfx |= RF_FULLBRIGHT;
	rocket->classname = "rocket";
	rocket->style = mod;

	if ( mod == MOD_ROCKET_S )	rocket->s.sound = trap_SoundIndex ( S_WEAPON_ROCKET_S_FLY );
	else						rocket->s.sound = trap_SoundIndex ( S_WEAPON_ROCKET_W_FLY );
}


//==================
//W_Touch_Plasma
//==================
void W_Touch_Plasma (edict_t *ent, edict_t *other, cplane_t *plane, int surfFlags)
{
	//don't hurt owner for the 1st second
	if (other == ent->r.owner) {
		if( !g_projectile_touch_owner->integer ||
			(g_projectile_touch_owner->integer && ent->timestamp + 1.0 > level.time) )
			return;
	}

	if (surfFlags & SURF_NOIMPACT)
	{
		G_FreeEdict (ent);
		return;
	}

	if( ent->style == MOD_PLASMA_S ) {

		T_RadiusDamage(ent, ent->r.owner, ent->dmg, ent->dmg_knockback, ent->radius_dmg, other, ent->dmg_radius, MOD_PLASMA_SPLASH_S);
		if (other->takedamage)
			T_Damage (other, ent, ent->r.owner, ent->velocity, ent->s.origin, plane->normal, ent->dmg, ent->dmg_knockback, 0, MOD_PLASMA_S);
		else {
			G_TurnEntityIntoEvent ( ent, EV_PLASMA_EXPLOSION, DirToByte (plane ? plane->normal : NULL) );
			ent->s.skinnum = FIRE_MODE_STRONG;
			return;
		}
	
	} else {

		T_RadiusDamage(ent, ent->r.owner, ent->dmg, ent->dmg_knockback, ent->radius_dmg, other, ent->dmg_radius, MOD_PLASMA_SPLASH_W);
		if (other->takedamage)
			T_Damage (other, ent, ent->r.owner, ent->velocity, ent->s.origin, plane->normal, ent->dmg, ent->dmg_knockback, 0, MOD_PLASMA_W);
		else {	
			G_TurnEntityIntoEvent ( ent, EV_PLASMA_EXPLOSION, DirToByte (plane ? plane->normal : NULL) );
			ent->s.skinnum = FIRE_MODE_WEAK;
			return;
		}
	}

	G_FreeEdict (ent);
}

//==================
//W_Fire_Plasma
//==================
void W_Fire_Plasma (edict_t *self, vec3_t start, vec3_t dir, int damage, int knockback, int radius_damage, int radius, int speed, int timeout, int mod )
{
	edict_t	*plasma;

	plasma = W_Fire_Projectile(self, start, dir, speed, damage, knockback, radius_damage, radius, timeout);
	plasma->s.modelindex = trap_ModelIndex ("models/objects/projectile/plasmagun/proj_plasmagun2.md3");
	plasma->touch = W_Touch_Plasma;
	plasma->classname = "plasma";
	plasma->s.renderfx |= RF_FULLBRIGHT;
	plasma->style = mod;

	if( mod == MOD_PLASMA_S )
		plasma->s.sound = trap_SoundIndex ( S_WEAPON_PLASMAGUN_S_FLY );
	else
		plasma->s.sound = trap_SoundIndex ( S_WEAPON_PLASMAGUN_W_FLY );
}

//==================
//W_Touch_Bolt
//==================
void W_Touch_Bolt( edict_t *self, edict_t *other, cplane_t *plane, int surfFlags )
{
	//don't hurt owner for the 1st second
	if (other == self->r.owner) {
		if( !g_projectile_touch_owner->integer ||
			(g_projectile_touch_owner->integer && self->timestamp + 1.0 > level.time) )
			return;
	}

	if (other->takedamage)
		T_Damage (other, self, self->r.owner, self->velocity, self->s.origin, plane->normal, self->dmg, self->dmg_knockback, 0, MOD_ELECTROBOLT_W);
	else 
	// add explosion event
	if( !(surfFlags & SURF_NOIMPACT) )
	{
		edict_t *event;
		event = G_SpawnEvent( EV_BOLT_EXPLOSION, DirToByte (plane ? plane->normal : NULL), self->s.origin );
		event->s.skinnum = FIRE_MODE_WEAK;
	}

	// free at next frame
	self->think = G_FreeEdict;
	self->touch = NULL;
#ifdef SV_FRAMETIME
	self->nextthink = level.timemsec + G_FRAMETIME;
#else
	self->nextthink = level.time + FRAMETIME;
#endif
}

//==================
//W_Fire_Electrobolt
//==================
void W_Fire_Electrobolt(edict_t *self, vec3_t start, vec3_t aimdir, float speed, int damage, int knockback, int timeout, int dflags, int mod)
{
	vec3_t	end;
	edict_t *event;
	edict_t	*bolt;
	trace_t		tr;

	// hitscan, strong mode
	if ( speed == 0.0 )
	{
		VectorMA( start, 4096, aimdir, end );
		trap_Trace (&tr, start, NULL, NULL, end, self, MASK_SHOT);

		// send gun puff / flash
		if (tr.fraction < 1.0)
		{
			if (game.edicts[tr.ent].takedamage)
			{
				T_Damage (&game.edicts[tr.ent], self, self, aimdir, tr.endpos, tr.plane.normal, damage, knockback, dflags, mod);
			}
			else
			{
				if ( !(tr.surfFlags & SURF_NOIMPACT) )
				{
				}
			}
		}

		// send the weapon fire effect
		event = G_SpawnEvent ( EV_ELECTROTRAIL, 0, start );
		event->r.svflags = SVF_FORCEOLDORIGIN;
		VectorCopy( tr.endpos, event->s.origin2 );
		event->s.ownerNum = self->s.number;//identify who fired for cgame
	}
	else
	{	// projectile, weak mode
		bolt = W_Fire_Projectile(self, start, aimdir, speed, damage, knockback, 0, 0, timeout);
		bolt->s.modelindex = trap_ModelIndex ("models/objects/projectile/electrobolt/proj_electrobolt.md3");
		bolt->s.type = ET_ELECTRO_WEAK; //add particle trail and light
		bolt->touch = W_Touch_Bolt;
		bolt->classname = "bolt";
		bolt->s.renderfx |= RF_FULLBRIGHT;
		bolt->style = mod;
	}
}

//==================
//W_Fire_Lasergun
//==================
void W_Fire_Lasergun( edict_t *self, vec3_t start, vec3_t aimdir, int damage, int knockback, int timeout, int dflags, int mod )
{
	int			i, playernum;
	edict_t		*e, *laser;
	trace_t		tr;
	vec3_t		end;

	// first of all, see if we already have a beam entity for this laser
	laser = NULL;
	playernum = ENTNUM(self);
	for( i = game.maxclients; i < game.maxentities; i++ )
	{
		e = &game.edicts[i];
		
		if( !e->r.inuse )
			continue;

		if( e->s.ownerNum == playernum && e->s.type == ET_LASERBEAM )
		{
			laser = e;
			G_Printf( "relink laser beam\n" );
			VectorCopy( laser->s.old_origin, laser->s.origin );
			break;
		}
	}

	// if no ent was found we have to create one
	if( laser == NULL )
	{
		G_Printf( "spawn laser beam\n" );
		laser = G_Spawn();
		laser->s.type = ET_LASERBEAM;
		laser->s.ownerNum = playernum;
		laser->movetype = MOVETYPE_NONE;
		laser->r.solid = SOLID_NOT;
		//laser->r.svflags = SVF_FORCEOLDORIGIN;
		laser->s.modelindex = trap_ModelIndex( "models/misc/beam_cross.md3" );
		//laser->s.skinnum = trap_ImageIndex( "gfx/misc/electro");
		//laser->s.modelindex = trap_ModelIndex( "models/objects/projectile/rlauncher/proj_rlauncher.md3" );
	}

	VectorMA( start, 1024, aimdir, end );
	trap_Trace (&tr, start, NULL, NULL, end, self, MASK_SHOT);
	// send gun puff / flash
	if (tr.fraction < 1.0)
	{
		if (game.edicts[tr.ent].takedamage) {
			T_Damage (&game.edicts[tr.ent], self, self, aimdir, tr.endpos, tr.plane.normal, damage, knockback, dflags, mod);
		} else {
			if ( !(tr.surfFlags & SURF_NOIMPACT) )
			{
			}
		}
	}

	//copy points
	VectorCopy( tr.endpos, laser->s.origin );
	//VectorCopy( start, laser->s.origin2 );
	

	// give it 2 server frames before freeing itself, so we can
	// relink it if we are still firing
	laser->think = G_FreeEdict;
#ifdef SV_FRAMETIME
	laser->nextthink = level.timemsec + ( G_FRAMETIME * 3 );
#else
	laser->nextthink = level.time + FRAMETIME * 3;
#endif

	trap_LinkEntity( laser );
}

//==============================================================================






